HOWTO
=====

This document provides all the necessary information required for seamless integration of new demo parts into `Homunculi` demo engine.

How to add a new part?
----------------------

Create a dedicated part directory. Comply with the naming conventions. Each individual part is added to a `src` subdirectory of a `Homunculi` project. Naming convention is required to be followed by each one of them. Part directory name consists of two digits indicating sequence number of a part in the course of a demo, followed by a dash, followed by a lowercased name without whitespaces (for the sake of simplicity of CLI operations executed in miscellaneous environments and operating systems).

The minimal setup requires to add the following project files:

* `Makefile` to define all build configuration targets
* `part-name.inc` to provide configurable start address of a part source code
* `part-name.src` to comprise a compilable and an executable source code

An include file must be named after a directory name containing the part code (following convention over configuration rule). It does not include a part sequence number however, only its name. Its compilation should not output any target file, it should serve only the purpose of collecting an entire part configuration in a single place. Therefore it needs to be included by the part source code file as well. Initial (starting) source code address label needs to follow a naming convention as well. It should match part name with all caps and dashes replaced by underscore characters, additionally followed by a `_START_ADDRESS` component. For a demo part named `fire-vector` this could for instance be:

    ;-------------------------------------------------------------------------------
    FIRE_VECTOR_START_ADDRESS = $4000
    ;-------------------------------------------------------------------------------

Here a starting address has been defined as `$4000` and this is were `exec_next_part` procedure will jump to when the time to execute this part comes (see section describing how to link to the next part for more details about `exec_next_part` procedure).

`part-name.src` source code file is compiled and loaded as a main executable component of any demo part. It needs to provide a startup procedure located at a previously defined start address.

    ;-------------------------------------------------------------------------------
    #include "part-name.inc"
    #include "../framework/framework.inc"
    #include "../runner/runner.inc"
    ;-------------------------------------------------------------------------------
            *= FIRE_VECTOR_START_ADDRESS
    ;-------------------------------------------------------------------------------
            ...
    ;-------------------------------------------------------------------------------

It should also take care of loading, decrunching, and executing a subsequent demo part (read the following sections for all the gory details on that). It should not define its own IRQ handler and setup procedures, rather refer to a predefined framework routines, which are always available during an entire course of the demo. Everything else is left entirely up to a programmer's preference.

Framework code provides a couple of useful routines available to run at any time. They include raster synchronisation, timer check, IRQ loader, ByteBoozer decruncher, music play, and IRQ setup amongst many others.

This is an example directory structure of an individual demo part, for instance named `fire-vector`:

    Homunculi
    `-- src
        `-- 06-fire-vector
            |-- fire-vector.inc
            |-- fire-vector.src
            `-- Makefile

All the irrelevant files have been left out of this picture in order not to obscure the clear overview of a minimal part setup. There is a sample part available in the source code repository, which can serve an empty template role for any of your demo parts. It is available as a `99-empty-template` directory and constitutes an excellent base and a starting point for further exploration of possible configuration ideas.

Once you are done with a basic setup, you are well prepared to begin writing your code. See below for a more detailed description of each required step and even more useful hints.

Memory requirements
-------------------

A memory area between the end of a soundtrack data up to the KERNAL IRQ vectors at `$fffa-$ffff` is at your disposal. As far as a zero page is concerned, you are advised to use anything between `$40-$ff` only, because lower zero page regions might be in use by IRQ loader, ByteBoozer decruncher, or demo framework procedures. Especially memory area between `$20-$3f` should be left untouched, because its modification might break an entire demo running engine.

How to document your part?
--------------------------

Because of tight interaction constraints between individual demo parts, it is important that neighbouring parts are well understood, so that they can be seamlessly integrated. Documenting your part carefully serves exactly this purpose.

Always provide an information about used and free memory in order to help figure out, where to load the next part into, while your part is still running. Provide information about how much memory does your part require to be loaded. This requires running a test build (an always healhty operation to assist validation of your setup) to figure out an exact size of a compressed target binary. Provide information where an initially uncompressed part code and data resides (before starting it) in a form of start and end addresses, so that previous part maintainer can place the code used to decompress and execute your part beyond these frames. It is perfectly fine to provide a documentation either in a form of a separate text file or simply include it directly in a `part-name.inc` part configuration file.

How to setup an IRQ routine?
----------------------------

In order to set up a new IRQ handler and apply it immediately, place the following bit of code into your program:

    ;-------------------------------------------------------------------------------
    #include "../framework/framework.inc"
    ;-------------------------------------------------------------------------------
            ...
    ;-------------------------------------------------------------------------------
            clc                ; ensure interrupt disable bit is cleared
            ldx #<irq          ; setup lower address byte of a custom IRQ handler
            ldy #>irq          ; setup higher address byte of a custom IRQ handler
            jsr irq_setup_addr ; call framework procedure to install new IRQ routine
    ;-------------------------------------------------------------------------------
            ...
    ;-------------------------------------------------------------------------------
    irq     jsr msx_play_exec  ; minimal IRQ handler should at least play demo music
                               ; (synchronisation counter is updated automatically)
            ldx #<irq          ; setup lower address byte of a next IRQ handler
            ldy #>irq          ; setup higher address byte of a next IRQ handler
            lda #$00           ; setup raster line which will trigger next interrupt
            rts                ; exit IRQ routine and leave control to the main code
    ;-------------------------------------------------------------------------------

Whenever timing matters, which is normally the case, as you may most likely want to execute your IRQ routine right at the beginning of a frame (raster line `$00`), setting interrupt disable status helps to avoid a premature interrupt. Using this method clearing interrupt disable bit can be delegated to a caller procedure:

    ;-------------------------------------------------------------------------------
    #include "../framework/framework.inc"
    ;-------------------------------------------------------------------------------
            ...
    ;-------------------------------------------------------------------------------
            sec                ; ensure interrupt disable status is set
            ldx #<irq          ; setup lower address byte of a custom IRQ handler
            ldy #>irq          ; setup higher address byte of a custom IRQ handler
            jsr irq_setup_addr ; call framework procedure to install new IRQ routine
            jsr raster_wait    ; wait for raster line to reach a bottom of a screen
            cli                ; make a new IRQ handler effective by enabling IRQs
    ;-------------------------------------------------------------------------------
            ...
    ;-------------------------------------------------------------------------------
    irq     jsr msx_play_exec  ; minimal IRQ handler should at least play demo music
                               ; (synchronisation counter is updated automatically)
            ldx #<irq          ; setup lower address byte of a next IRQ handler
            ldy #>irq          ; setup higher address byte of a next IRQ handler
            lda #$00           ; setup raster line which will trigger next interrupt
            rts                ; exit IRQ routine and leave control to the main code
    ;-------------------------------------------------------------------------------

Before leaving your part (and moving on to the next demo part), it is usually recommended to reset current IRQ routine with a default one. A default IRQ routine performs a very basic procedure which plays music and increments synchronisation counter respectively. If your part is supposed to perform a more sophisticated code upon exit, this is not a recommended solution, and thus you should still attempt to write your own incorruptible routine. In order to set a default IRQ handler, add the following bit of code into your program:

    ;-------------------------------------------------------------------------------
    #include "../framework/framework.inc"
    ;-------------------------------------------------------------------------------
            ...
    ;-------------------------------------------------------------------------------
            clc                 ; ensure interrupt disable bit is cleared
            jsr irq_set_default ; set default IRQ which cannot be overwritten
    ;-------------------------------------------------------------------------------

Setting up a default IRQ routine ensures that each subsequent interrupt will be triggered at a top of a screen (raster line `$00`) until explicitly changed.

Please do not customize IRQ setup yourself, as this might break an entire demo engine. Always use provided framework routines!

How to link to the next part?
-----------------------------

It is responsibility of yours to ensure proper loading, depacking, and executing of a next part of the demo. Please refer to a documentation of a particular demo part to figure out when and where each step of a linking procedure may be performed. It is reasonable to locate your linking code in a memory region that cannot be overwritten by unpacking the next part, while at the same time making sure that this region does not overwrite the linking code of a previous part (such an accurate synchronisation requires careful planning and is best achieved by an easily configurable and properly documented source code start addresses, so that they can be changed seamlessly).

Your part is going to be automatically linked and executed by a preceding part, so you can forthwith test what are the current values of a `SYNCHRO_COUNTER` (see `framework.inc` for more details) during your part's execution in order to setup proper timings. First thing you are required to do is an incrementation of a global part index. Once this is accomplished, you should pick the right moment to load and depack the next part by calling appropriate framework routines. As mentioned in a previous section, you might want to switch to a default IRQ routine once your part is completed (please note that the default IRQ handler does not perform anything besides playing music). It is perfectly fine to keep your own IRQ procedure running during decrunching and initial execution of the next part (until its setup switches to an own IRQ routine), as long as you are able to ensure the code won't be overwritten under any circumstances. This has an advantage of a smooth transition between individual parts, because your own code can be directly taken over by the next part while it is still running. If that is the case with a previous part of yours, you should make sure to begin with a fade out routine that elegantly blanks the screen before initialising your own part.

    ;-------------------------------------------------------------------------------
    #include "../framework/framework.inc"
    #include "../runner/runner.inc"
    ;-------------------------------------------------------------------------------
    NEXT_PART_START_ADDRESS     = $a000 ; safe address that won't be overwritten
    ;-------------------------------------------------------------------------------
    READY_TO_LOAD_NEXT_PART     = $3120 ; timestamp indicating readiness to load
    READY_TO_DECRUNCH_NEXT_PART = $3160 ; timestamp indicating readiness to decrunch
    ;-------------------------------------------------------------------------------
            *= NEXT_PART_START_ADDRESS
    ;-------------------------------------------------------------------------------
            ; Increment the demo part counter:
            inc NEXT_PART_INDEX

            ; Load the next demo part from disk:
            ldx #<READY_TO_LOAD_NEXT_PART
            ldy #>READY_TO_LOAD_NEXT_PART
            jsr synchro_check
            jsr load_next_part

            ; Set default IRQ, otherwise new part would have overwritten IRQ:
            ldx #<READY_TO_DECRUNCH_NEXT_PART
            ldy #>READY_TO_DECRUNCH_NEXT_PART
            jsr synchro_check
            clc
            jsr irq_set_default

            ; Decrunch the next demo part:
            jsr decr_next_part

            ; Execute the next demo part:
            jmp exec_next_part
    ;-------------------------------------------------------------------------------

How to configure build process?
-------------------------------

Build process of each individual demo part needs to follow a couple of rules. An entire build system is based upon a coupled suite of makefiles, thus every part is required to provide a `Makefile` with at least the following targets: `$(TARGET_DISK_FILES)`, `$(TARGET_DIR)`, `clean`, and `distclean`. For convenience of a build flexibility and robustness it is highly recommendeded to setup a few additional targets as well: `$(BUILD)`, `$(COMPILE)`, `$(SOURCE_CODE)`. A `default` target should normally default to `$(TARGET_DISK_FILES)`. All of the aforementioned targets are fully described below:

* `$(TARGET_DIR)` provides means of creation of a conventional build directory
* `$(SOURCE_CODE)` ensured an earlier creation of a build directory for all compiled files
* `$(COMPILE)` compiles an entire part source code
* `$(BUILD)` uses a default crunching procedure to compress and relocate a compiled binary
* `$(TARGET_DISK_FILES)` is used to copy over all part files into a temporary directory which is later used as a source to setup target demo disk image
* `clean` deletes all target disk files, and both compiled and crunched binaries
* `distclean` removes all precompiled dependencies, all compiled sources, and all target objects from disk

Additionally, part file name on a demo disk and its loading address should not be hard-coded into a `Makefile`, since a couple of auxiliary shell functions has been provided to retrieve their values:

    PART_NAME=$(shell source ../../common/func.sh; part_file_name)
    LOADING_ADDRESS=$(shell source ../../common/func.sh; loading_address)

Using this approach we get more robust and more flexible configuration resilience. Loading address is a memory address, where the final target binary will be relocated to in order to allow its loading during previous part execution (it is configurable by a `load_addr` property of a demo part configuration defined in a `src/config.yml` file). Make sure to pick a reasonable loading address for your part. For example if your part occupies `$0800` bytes of a memory after compression, the best choice for a loading address will most likely be `$f600`.

This setup imposes a couple of constraints on your part, amongst which the most important ones are:

* Your source code must compile, thus needs to be implemented in a `dreamass`-friendly syntax
* Any time you add or remove files from/to your project, you should make sure that prerequisites are always complete and up-to-date
* Only source code should be checked into the `git` repository, and all binaries should be constructable by appropriate `make` tasks

An example `Makefile` designed for a feigned demo part named `fire-vector` is presented below:

    include ../../common/config

    FRAMEWORK=../framework/framework.inc
    PREREQUISITES=fire-vector.inc

    SOURCE_CODE=fire-vector.src
    COMPILE=$(TARGET_DIR)/fire-vector.prg
    BUILD=$(TARGET_DIR)/fire-vector

    PART_NAME=$(shell source ../../common/func.sh; part_file_name)
    LOADING_ADDRESS=$(shell source ../../common/func.sh; loading_address)

    default: $(TARGET_DISK_FILES)

    $(TARGET_DIR):
        mkdir -vp $@

    $(SOURCE_CODE): $(TARGET_DIR)

    $(COMPILE): $(SOURCE_CODE) $(FRAMEWORK) $(PREREQUISITES)
        $(DREAMASS) -v $< -o $@

    $(BUILD): $(COMPILE)
        ../../tools/scripts/crunch.pl $< $@ $(LOADING_ADDRESS)

    $(TARGET_DISK_FILES): $(TARGET_DIR) $(BUILD)
        rm -rf $@
        mkdir -vp $@
        cp -v $(BUILD) $@/$(PART_NAME)

    clean:
        rm -rf $(TARGET_DISK_FILES)
        $(RM) $(BUILD)
        $(RM) $(COMPILE)

    distclean: clean
        rm -rf $(TARGET_DIR)

How to add new part to demo configuration?
------------------------------------------

Demo configuration file is located in the `src` subdirectory of a `Homunculi` project and is named `config.yml`. It is a YAML-formatted text file consisting of a sorted list of all demo parts. Each list item is a hash with one key with a corresponding subdirectory name containing all of the part code, and a value to this key, which is another hash containing two greatly important properties: part relocation/loading address `load_addr` and synchronisation timer `sync_time`.

* `load_addr` defines loading address for a compressed binary of your part code (which is loaded by a runner's `load_next_part` and utilised by `decr_next_part` in order to correctly decrunch it)
* `sync_time` provides a timing information that is utilised by a runner's `exec_next_part` routine (when your part is executed, it loops infinitely until synchronisation time you provided is reached by a `SYNCHRO_COUNTER`)

Here is an example of a feigned demo part named `fire-vector`, how it could be configured:

    - 06-fire-vector:

        load_addr: $ea00
        sync_time: $2b70

How to compile and execute a single part?
-----------------------------------------

Although compilation and execution of an individual part of a demo is not implemented yet, you are able to begin a demo run at a selected part, which pretty much saves you from all the looping over preceding parts any time you just want to test a specific part. When compiling a demo you simply pass an additional parameter named `PART`, which expects a two-digit decimal number as its value:

    $ make PART=07
    $ make test PART=07

This command will fast-forward the demo up to the part number `07`, which begins the second part of a presentation (`Winter Games`).

What if my part loads additional files?
---------------------------------------

If your part loads additional files, make sure you build them using an adequate `make` task and include them into `$(TARGET_DISK_FILES)` target, which copies them over into a temporary directory used as a source for building target demo disk image. All the other operations will be automatically handled by the demo engine. Here is an example of a `Makefile` configuration providing all the necessary setup for adding a new part file to our good friend `fire-vector`:

    ADDITIONAL_FILE=$(TARGET_DIR)/ab

    PREREQUISITES=fire-vector.inc $(ADDITIONAL_FILE)

    $(ADDITIONAL_FILE): $(TARGET_DIR)
        scripts/build-additional-file.pl --target-file=$(ADDITIONAL_FILE)

    $(TARGET_DISK_FILES): $(TARGET_DIR) $(BUILD)
        rm -rf $@
        mkdir -vp $@
        cp -v $(BUILD) $@/$(PART_NAME)
        cp -v $(ADDITIONAL_FILE) $@/

You can load and decrunch any of your files, using general-purpose framework routines, like in the following example:

    ;-------------------------------------------------------------------------------
    #include "../framework/framework.inc"
    ;-------------------------------------------------------------------------------
    LOADING_ADDRESS = $b800 ; loading address of a decrunched file
    ;-------------------------------------------------------------------------------
            ...
    ;-------------------------------------------------------------------------------
            clc                   ; do not decompress file on the fly
            ldx #<fname           ; setup lower filename address vector
            ldy #>fname           ; setup higher filename address vector
            jsr loadraw           ; load file from disk

            ldy #<LOADING_ADDRESS ; setup lower loading address vector
            ldx #>LOADING_ADDRESS ; setup higher loading address vector
            jsr decrunch          ; decrunch recently loaded file
    ;-------------------------------------------------------------------------------
            ...
    ;-------------------------------------------------------------------------------
    fname   .dp "ab"
            .db $00
    ;-------------------------------------------------------------------------------

There is one caveat however. You need to make sure your IRQ routine handler is not pointing to the code in the area of $d000-$ffff, because the demo engine turns off RAM above $d000 when loading process is in progress for IDE64 compatibility reasons. It is fine to access RAM above $e000 within your IRQ routine though, because a wrapper code for your IRQ handler switches KERNAL off for the time being of your IRQ routine execution.

How to synchronise code with a music timer?
-------------------------------------------

Outside an IRQ routine when you want to wait until a certain timer value is eventually reached, use a blocking version of a check procedure:

    ;-------------------------------------------------------------------------------
    #include "../framework/framework.inc"
    ;-------------------------------------------------------------------------------
    SYNCHRO_TIMESTAMP = $02c0
    ;-------------------------------------------------------------------------------
            ...
    ;-------------------------------------------------------------------------------
            ldx #<SYNCHRO_TIMESTAMP
            ldy #>SYNCHRO_TIMESTAMP
            jsr synchro_check
    ;-------------------------------------------------------------------------------

Synchronisation check will loop until:

  * .X/.Y is greater than or equal to `SYNCHRO_COUNTER` (meaning that synchronisation counter has effectively been reached)

Inside an IRQ routine, when you only want to check whether a certain timer value has been already reached, you cannot block code execution, so you have to use a non-blocking version of a check procedure:

    ;-------------------------------------------------------------------------------
    #include "../framework/framework.inc"
    ;-------------------------------------------------------------------------------
    SYNCHRO_TIMESTAMP = $02c0
    ;-------------------------------------------------------------------------------
            ...
    ;-------------------------------------------------------------------------------
            ldx #<SYNCHRO_TIMESTAMP
            ldy #>SYNCHRO_TIMESTAMP
            jsr synchro_check_no_wait
            bcs timestamp_reached
    ;-------------------------------------------------------------------------------

This synchronisation check will:

  * Set carry when .X/.Y is greater than or equal to `SYNCHRO_COUNTER` (meaning that synchronisation counter has effectively been reached)
  * Clear carry when .X/.Y is less than `SYNCHRO_COUNTER` (meaning that synchronisation counter has not yet been reached)
