#!/usr/bin/perl
#
# Dump source data:
#
# perl twister1.pl
#
# Animate prototype:
#
# perl twister1.pl --animate

use utf8;
use bytes;
use strict;
use warnings;

use Data::Dumper;
use Getopt::Long;
use Source::Data qw(generate_source_data);
use Tk;

our $width = 320;
our $height = 200;
our $bckgrd = '#CCCCCC';
our $colour1 = '#004400';
our $colour2 = '#44cc44';
our $num_steps = 256;
our $max_value = 32;
our $general_step = $num_steps / 4;
our $twister_effect_background_colour = 0x0c;

# Get options:
my %options = get_options();

our $sinus = calculate_sinus();
our $steps = calculate_steps($sinus);
# check_duplicates($steps);
# warn Dumper $steps;
generate_tables($steps);

# Animate prototype when "animate" option has been specified:
if ($options{animate}) {

  our $mw = MainWindow->new;

  our $canvas = $mw->Canvas(-width => 2 * $width, -height => 2 * $height, -background => $bckgrd)->pack;

  $mw->after(1 => sub { do_draw(0, $mw, $canvas) });

  MainLoop;
}

sub do_draw {
  my ($step, $mw, $canvas) = @_;
  $canvas->delete('all');

  for (my $y = 0; $y < 25; $y++) {

    my $index = ($step + ($y / 2)) & (scalar (@{$steps}) - 1);

    my $step = $steps->[$index];

    for my $part (@{$step}) {

      my $from = ($part->[0] + 0x30) * 4;
      my $to = ($part->[1] + 0x30) * 4;
      my $colour = $part->[2];

      for (my $j = 0; $j < 8; $j++) {

        my $y1 = 2 * (8 * $y + $j);
        my $y2 = $y1 + 1;

        $canvas->createLine($from, $y1, $to, $y1, -width => 1, -fill => $colour);
        $canvas->createLine($from, $y2, $to, $y2, -width => 1, -fill => $colour);
      }
    }
  }

  ## Next frame
  $mw->after(10 => sub { do_draw(++$step, $mw, $canvas) });
}

sub generate_tables {
  my ($steps) = @_;

  my @colors;

  for my $step (@{$steps}) {

    my @step_colors;

    my $initial_part = 1;

    for my $part (@{$step}) {

      my $from = int ($part->[0] / 2);
      my $to = int ($part->[1] / 2);
      my $colour = $part->[2] eq $colour1 ? 0x05 : 0x0d;

      if ($initial_part) {
        for (0x00 .. $from - 0x01) {
          push @step_colors, $twister_effect_background_colour;
        }
      }

      for ($from .. $to - 0x01) {
        push @step_colors, $colour;
      }

      $initial_part = 0;
    }

    push @colors, \@step_colors;
  }

  my $max_values_in_entry = [ reverse sort map { scalar @{$_} } @colors ]->[0];

  @colors = map {
    my @color = @{$_};
    for (scalar (@color) .. $max_values_in_entry) {
      push @color, $twister_effect_background_colour;
    }
    \@color;
  } @colors;

  my @source_data;

  my $i = 1;

  for my $color (@colors) {

    my @single_colors = @{$color};

    my @merged_colors;

    while (@single_colors) {

      my $merged_color = sprintf '$%x%x', splice @single_colors, 0x00, 0x02;
      push @merged_colors, $merged_color;
    }

    push @source_data, {
      data => \@merged_colors,
      name => sprintf ('twister_data_index_%02d', $i++),
    };
  }

  dump_source_data(\@source_data);

  return;
}

sub dump_source_data {
  my ($data) = @_;

  # Dump lo/hi indexes lookup tables:
  my @index_names = map { $_->{name} } @{$data};

  my $lo_hi_indexes_source_code = generate_source_data(
    data                   => \@index_names,
    data_type              => 'bytes',
    expected_indentation   => 20,
    include_comments       => 1,
    lo_and_hi_tables_gen   => 1,
    name                   => 'twister_data_index',
    values_per_line        => 1,
  );

  print $lo_hi_indexes_source_code;

  print "\n";

  # Dump actual data indexes lookup tables:
  for (my $i = 0; $i < @{$data}; $i++) {

    my $index = $data->[$i];

    my $index_data = $index->{data};
    my $index_name = $index->{name};

    my $index_data_source_code = generate_source_data(
      data                   => $index_data,
      data_type              => 'bytes',
      expected_indentation   => 20,
      include_comments       => $i == 0,
      lo_and_hi_tables_gen   => 0,
      name                   => $index_name,
      values_per_line        => 16,
    );

    print $index_data_source_code;
  }

  return;
}

sub check_duplicates {
  my ($steps) = @_;

  my %elements;

  my $i = 0;

  for my $step (@$steps) {

    my $token = join '/', map {
      my ($from, $to, $colour) = @{$_};
      sprintf '%02x-%02x-%s', $from, $to, $colour;
    } @{$step};

    $elements{$token} .= $i++ . ' ';
  }

  warn Dumper \%elements;

  return;
}

sub calculate_steps {
  my ($sinus) = @_;

  my @steps;

  for (my $index = 0; $index < $num_steps / 8; $index++) {

    my $x1 = $sinus->[0]->[$index];
    my $x2 = $sinus->[1]->[$index];
    my $x3 = $sinus->[2]->[$index];
    my $x4 = $sinus->[3]->[$index];

    my @parts;

    _push_part(\@parts, $x1, $x2, $colour2);
    _push_part(\@parts, $x2, $x3, $colour1);
    _push_part(\@parts, $x3, $x4, $colour2);
    _push_part(\@parts, $x4, $x1, $colour1);

    @parts = sort { $a->[0] > $b->[0] } @parts;

    push @steps, \@parts;
  }

  return \@steps;
}

sub _push_part {
  my ($parts, $from, $to, $colour) = @_;

  if ($from < $to) {
    push @{$parts}, [$from / 4 - 0x30, $to / 4 - 0x30, $colour];
  }

  return;
}

sub calculate_sinus {
  my @sinus = ([], [], [], []);

  my @przes = (0, 3.14 / 2, 3.14, 3 * 3.14 / 2);

  for (my $i = 0; $i < @przes; $i++) {

    my $przes = $przes[$i];

    for (my $j = 0; $j < $num_steps; $j++) {
      my $sinus = 4 * int ($max_value * sin(4 * 2 * 3.14 * $j / $num_steps + $przes)) + $width;
      push @{$sinus[$i]}, $sinus;
    }
  }

  return \@sinus;
}

sub get_options {
  my %options = (
    animate => 0,
  );

  GetOptions ("animate" => \$options{animate})
  or die("Error in command line arguments\n");

  return %options;
}
