#!/usr/bin/perl

=head1 NAME

create-compressed-image-cuts.pl - Create configured extractions of scrolled FacePainter image(s) and compress all generated files with ByteBoozer

=head1 SYNOPSIS

  ./create-compressed-image-cuts.pl --target-directory=<target-directory> --target-config=<target-config>

=head1 EXAMPLES

  ./create-compressed-image-cuts.pl --target-directory=../image-cuts --target-config=../image-cuts-config.src

=cut

use utf8;
use bytes;
use strict;
use warnings;

local $| = 1;

use lib q{../../tools/lib};

use Archive::ByteBoozer qw/crunch/;
use Data::Dumper;
use Getopt::Long;
use IO::File;
use Memory::Address qw(:file :format);
use Source::Data qw(generate_source_data);
show_headline();
my %options = get_options();

printf qq{Creating "%s" target directory to store all generated files...}, $options{target_directory};
mkdir $options{target_directory};
print " OK!\n\n";

my %config = (
  target_names_scheme => 's%d.prg',
  files => [
    {
      # Will be uncompressed to: $6f00-$8288:
      'filename'                   => 'images/01.koa',
      'offset'                     => 0,
      'length'                     => 20,
      'replace_loading_address'    => 0x6f00,
      'shift_relocated_data_up_to' => 0x8300,
    },
    {
      # Will be uncompressed to: $e000-$f388:
      'filename'                   => 'images/01.koa',
      'offset'                     => 20,
      'length'                     => 20,
      'replace_loading_address'    => 0xe000,
      'shift_relocated_data_up_to' => 0xf400,
    },
    {
      # Will be uncompressed to: $6f00-$8288:
      'filename'                   => 'images/02.koa',
      'offset'                     => 0,
      'length'                     => 20,
      'replace_loading_address'    => 0x6f00,
      'shift_relocated_data_up_to' => 0x8300,
    },
    {
      # Will be uncompressed to: $e000-$f388:
      'filename'                   => 'images/02.koa',
      'offset'                     => 20,
      'length'                     => 20,
      'replace_loading_address'    => 0xe000,
      'shift_relocated_data_up_to' => 0xf400,
    },
    {
      # Will be uncompressed to: $6f00-$8288:
      'filename'                   => 'images/03.koa',
      'offset'                     => 0,
      'length'                     => 20,
      'replace_loading_address'    => 0x6f00,
      'shift_relocated_data_up_to' => 0x8300,
    },
    {
      # Will be uncompressed to: $e000-$f388:
      'filename'                   => 'images/03.koa',
      'offset'                     => 20,
      'length'                     => 20,
      'replace_loading_address'    => 0xe000,
      'shift_relocated_data_up_to' => 0xf400,
    },
    {
      # Will be uncompressed to: $6f00-$8288:
      'filename'                   => 'images/04.koa',
      'offset'                     => 0,
      'length'                     => 20,
      'replace_loading_address'    => 0x6f00,
      'shift_relocated_data_up_to' => 0x8300,
    },
    {
      # Will be uncompressed to: $e000-$f388:
      'filename'                   => 'images/04.koa',
      'offset'                     => 20,
      'length'                     => 20,
      'replace_loading_address'    => 0xe000,
      'shift_relocated_data_up_to' => 0xf400,
    },
  ],
);

my $target_names_scheme = $config{target_names_scheme};

print "Extracting scrolled FacePainter images and compressing all generated files with ByteBoozer...\n\n";

my @files = @{$config{files}};
for (my $i = 0; $i < @files; $i++) {
  my $file = $files[$i];
  my $filename = $file->{'filename'};
  my $offset = $file->{'offset'};
  my $length = $file->{'length'};
  my $replace_loading_address = $file->{'replace_loading_address'};
  my $shift_relocated_data_up_to = $file->{'shift_relocated_data_up_to'};

  my $target_filename = sprintf "%s/$target_names_scheme", $options{target_directory}, $i + 1;

  my $command = sprintf 'scripts/extract-image-column-data.pl -s %s -t %s -o %d -l %d -a %d', $filename, $target_filename, $offset, $length, $replace_loading_address;
  printf "\$ %s\n", $command;
  system $command;

  (my $compressed_file = $target_filename) =~ s/\.prg$//;
  $file->{compressed_filename} = $compressed_file;

  printf qq{Compressing extracted "%s" image file with ByteBoozer into "%s" output file...\n\n}, $target_filename, $compressed_file;

  my $original = new IO::File $target_filename, 'r';
  my $crunched = new IO::File $compressed_file, 'w';
  crunch(source => $original, target => $crunched, relocate_output_up_to => $shift_relocated_data_up_to, verbose => 1);

  print "\n";
}

printf qq{Generating "%s" source code file with extracted image filenames and loading addresses...}, $options{target_config};

my %source_data;
for my $file (@{$config{files}}) {
  my $compressed_file = $file->{compressed_filename};
  (my $filename = $compressed_file) =~ s!^.*/!!;

  my $loading_address = address_to_hex_string(get_file_loading_address($compressed_file));

  push @{$source_data{loading_address}}, $loading_address;

  my $first_letter = substr $filename, 0, 1;
  my $second_letter = substr $filename, 1, 1;

  push @{$source_data{first_letter}}, $first_letter;
  push @{$source_data{second_letter}}, $second_letter;
}

my $source_code;
while (my ($key, $data) = each %source_data) {
  my $label = sprintf 'scrolled_image_%s', $key;

  my $key_source_code = generate_source_data(
    additional_indentation => 32,
    data                   => $data,
    data_type              => $key eq 'loading_address' ? 'bytes' : 'petscii',
    expected_indentation   => 34,
    include_comments       => 1,
    lo_and_hi_tables_gen   => $key eq 'loading_address' ? 1 : 0,
    name                   => $label,
    values_per_line        => $key eq 'loading_address' ? 4 : 8,
  );

  $source_code .= "$key_source_code\n";
}

my $fh = new IO::File $options{target_config}, 'w';
if (defined $fh) {
  $fh->binmode(':utf8');
  print $fh $source_code;
  $fh->close;
}

print " OK!\n\n";

sub show_headline {
    print <<HEADLINE;

Create Compressed Image Cuts 0.01 (2013-11-17)
Copyright (C) 2013 Pawel Krol (DJ Gruby/Arsenic)

HEADLINE

    return;
}

sub get_options {
  my ($help, $target_directory, $target_config);

  GetOptions(
    'help|?'               => \$help,
    'target-directory|d=s' => \$target_directory,
    'target-config|c=s'    => \$target_config,
  );

  if (defined $help) {
    show_usage->();
    exit 0;
  }

  unless (defined $target_directory) {
    show_usage->(qq{Error: missing argument: <target-directory>});
    exit 1;
  }

  if (-e $target_directory) {
    show_usage->(qq{Error: target directory already exists: "$target_directory"});
    exit 2;
  }

  unless (defined $target_config) {
    show_usage->(qq{Error: missing argument: <target-config>});
    exit 3;
  }

  if (-e $target_config) {
    show_usage->(qq{Error: target config file already exists: "$target_config"});
    exit 4;
  }

  return (
    target_config    => $target_config,
    target_directory => $target_directory,
  );
}

sub show_usage {
    my ($error) = @_;

    print qq{$error\n\n} if $error;

    print <<USAGE;
Usage: $0 --target-directory=<target-directory> --target-config=<target-config>

    <target-directory>
        Target directory to store all generated files (e.g., "image-cuts")
    <target-config>
        Target source code config file with all loading addresses (e.g., "image-cuts-config.src")

USAGE

    return;
}
