#!/usr/bin/perl

=head1 NAME

extract-image-column-data.pl - Extract out any number of columns out of a KoalaPainter image

=head1 SYNOPSIS

  ./extract-image-column-data.pl --source=<source> --target=<target> --offset=<offset> --length=<length> --start-address=<start-address>

=head1 EXAMPLES

  ./extract-image-column-data.pl --source=frighthof83-yazoo.koa --target=01.prg --offset=20 --length=10 --start-address=57344

=cut

use utf8;
use bytes;
use strict;
use warnings;

local $| = 1;

use lib q{../../tools/lib};

use Data::Dumper;
use Getopt::Long;
use Graphics::KoalaPainter;
use IO::File;
use Memory::Address qw(:all);

show_headline();
my %options = get_options();

printf qq{Extracing %d columns of image data at %d column offset from "%s" file...}, $options{'length'}, $options{'offset'}, $options{'source'};

my $picture = Graphics::KoalaPainter->new(file => $options{'source'});

my $from = $options{'offset'};
my $to = $options{'offset'} + $options{'length'} - 1;
my $columns = $picture->get_columns_data(from => $from, to => $to);

print " OK!\n\n";

my $address = address_to_lo_hi_bytes($options{'start_address'});

my $data = $address . $columns->{bitmap} . $columns->{screen} . $columns->{colors};

printf qq{Writing %d bytes of extracted image column data into "%s" file...}, length ($data), $options{target};

my $fh = new IO::File $options{target}, 'w';
if (defined $fh) {
  $fh->binmode(':bytes');
  syswrite $fh, $data;
  $fh->close;
}

print " OK!\n\n";

sub show_headline {
    print <<HEADLINE;

Extract KoalaPainter Image Column Data 0.01 (2013-11-17)
Copyright (C) 2013 Pawel Krol (DJ Gruby/Arsenic)

HEADLINE

    return;
}

sub get_options {
  my ($help, $source, $target, $offset, $length, $start_address);

  GetOptions(
    'help|?'            => \$help,
    'source|s=s'        => \$source,
    'target|t=s'        => \$target,
    'offset|o=i'        => \$offset,
    'length|l=i'        => \$length,
    'start-address|a=i' => \$start_address,
  );

  if (defined $help) {
    show_usage->();
    exit 0;
  }

  unless (defined $source) {
    show_usage->(qq{Error: missing argument: <source>});
    exit 1;
  }

  unless (-e $source) {
    show_usage->(qq{Error: file does not exist: "$source"});
    exit 2;
  }

  unless (defined $target) {
    show_usage->(qq{Error: missing argument: <target>});
    exit 3;
  }

  if (-e $target) {
    show_usage->(qq{Error: file already exists: "$target"});
    exit 4;
  }

  unless (defined $offset) {
    show_usage->(qq{Error: missing argument: <offset>});
    exit 5;
  }

  unless ($offset =~ m/^\d+$/ && $offset >= 0 && $offset <= 39) {
    show_usage->(qq{Error: invalid offset value: "$offset"});
    exit 6;
  }

  unless (defined $length) {
    show_usage->(qq{Error: missing argument: <length>});
    exit 7;
  }

  unless ($length =~ m/^\d+$/ && $length >= 1 && $length <= 40) {
    show_usage->(qq{Error: invalid length value: "$length"});
    exit 8;
  }

  unless (defined $start_address) {
    show_usage->(qq{Error: missing argument: <start-address>});
    exit 9;
  }

  unless ($start_address =~ m/^\d+$/ && $start_address >= 0 && $start_address <= 65535) {
    show_usage->(qq{Error: invalid start address value: "$start_address"});
    exit 8;
  }

  return (
    'source'        => $source,
    'target'        => $target,
    'length'        => $length,
    'offset'        => $offset,
    'start_address' => $start_address,
  );
}

sub show_usage {
    my ($error) = @_;

    print qq{$error\n\n} if $error;

    print <<USAGE;
Usage: $0 --source=<source> --target=<target> --offset=<offset> --length=<length> --start-address=<start-address>

    <source>
        Original KoalaPainter filename (e.g., "frighthof83-yazoo.koa")
    <target>
        Generated binary data filename (e.g., "01.prg")
    <offset>
        First extracted column index, must be between 0 and 39 (e.g., "20")
    <length>
        Number of extracted columns, must be between 1 and 40 (e.g., "10")
    <start-address>
        Start address of a generated binary data file (e.g., "57344")

USAGE

    return;
}
