package Vector1;

use bytes;
use strict;
use utf8;
use warnings;

use Data::Dumper;
use List::MoreUtils qw(uniq);
use Readonly;
use Vector3D::Easy qw(:all);

Readonly our $pi        => 3.14;
Readonly our $eyes      => [0, 0, 600];
Readonly our $num_steps => 128;
Readonly our $scale     => 17.3;
Readonly our $translate_x => 31;
Readonly our $translate_y => 31;
Readonly our $divide_x    => 2;

sub new {
  my ($this, %args) = @_;
  my $class = ref $this || $this;
  my $self = {};
  bless $self, $class;
  $self->_initialize(%args);
  return $self;
}

sub get_steps {
  my ($self) = @_;

  # Fetch stored steps to avoid duplicate calculations:
  my $steps = $self->{steps};
  if (defined $steps) {
    return @{$steps};
  }

  my @steps = $self->_get_steps_as_c64_lines();
  $self->_rearrange_steps_by_colours(\@steps);

  # warn Dumper $self->check_min_max_coords(\@steps);

  # Store computed steps to avoid duplicate calculations:
  $self->{steps} = \@steps;

  return @steps;
}

sub build_lines {
  my ($self, $steps) = @_;

  my @steps = defined $steps ? @{$steps} : $self->get_steps();

  my @lines;

  for (my $i = 0; $i < @steps; $i++) {

    my $step = $steps[$i];

    my $line = $self->_step_to_lines($step, $i);

    push @lines, $line;
  }

  return @lines;
}

sub get_lines_source_code {
  my ($self, $lines) = @_;

  my @lines = defined $lines ? @{$lines} : $self->build_lines();

  my $source_code;

  $source_code .= sprintf "NUM_STEPS_1 = %d\n", scalar @lines;

  my $i = 1;

  for my $frame (@lines) {

    $source_code .= sprintf "\n        ; Frame #%d\n", $i++;

    my $j = 1;

    my $colours = $self->{colours};

    for my $colour (@{$colours}) {

      my @colour_lines = @{$frame->{$colour}};

      $source_code .= sprintf "        .db \$%02x ; number of all values for colour #%d\n", scalar (@colour_lines) * 7, $j++;

      for my $colour_line (@colour_lines) {

        $source_code .= sprintf "        .db %s\n", join ',', map { sprintf '$%02x', $_ } @{$colour_line};
      }
    }
  }

  return $source_code;
}

sub check_min_max_coords {
  my ($self, $steps) = @_;

  my $any_point = [ map { @{$_} } values %{$steps->[0]} ]->[0];

  my %min_max_coords = (
    min_x => $any_point->[0],
    max_x => $any_point->[0],
    min_y => $any_point->[1],
    max_y => $any_point->[1],
  );

  for my $step (@{$steps}) {
    for my $polygon (values %{$step}) {
      for my $point (@{$polygon}) {

        my $x = $point->[0];
        my $y = $point->[1];

        $min_max_coords{min_x} = $x if $x < $min_max_coords{min_x};
        $min_max_coords{max_x} = $x if $x > $min_max_coords{max_x};
        $min_max_coords{min_y} = $y if $y < $min_max_coords{min_y};
        $min_max_coords{max_y} = $y if $y > $min_max_coords{max_y};
      }
    }
  }

  return \%min_max_coords;
}

sub speed {
  my ($self) = @_;

  my $speed = $num_steps / 4;

  return $speed;
}

sub next {
  my ($self) = @_;

  my $plane = $self->{'next'}->();

  return $plane;
}

sub _initialize {
  my ($self, %args) = @_;

  $self->{width}    = $args{width};
  $self->{height}   = $args{height};
  $self->{points}   = $args{points};
  $self->{polygons} = $args{polygons};
  $self->{colours}  = $args{colours};

  my @steps = $self->_setup_steps();

  # frame -> plane -> vertex -> x,y
  my @planes = $self->_setup_planes(@steps);
  $self->{planes} = \@planes;

  my $step = 0;

  $self->{'next'} = sub {
    my $next = $planes[$step++ & ($num_steps - 1)];
    return $next;
  };

  return;
}

sub _setup_steps {
  my ($self) = @_;

  my $width  = $self->{width};
  my $height = $self->{height};
  my $points = $self->{points};

  my @steps;

  for (my $i = 0; $i < $num_steps; $i++) {

    my $rotate_x = - 4 * $pi * $i / $num_steps;
    my $rotate_y = 2 * $pi * $i / $num_steps;
    my $rotate_z = 2 * $pi * $i / $num_steps;

    # 1. Rotate
    # 2. Scale
    # 3. Cast
    # 4. Translate
    my @step = map {
      translate_point($_, [ $width, $height, 0 ] ),
    } map {
      perspective_cast($_, $eyes),
    } map {
      scale_point($_, $scale)
    } map {
      rotate_point($_, [$rotate_x, $rotate_y, $rotate_z])
    } @{$points};

    push @steps, \@step;
  }

  return @steps;
}

sub _setup_planes {
  my ($self, @steps) = @_;

  my $polygons = $self->{polygons};

  my @planes;

  for my $step (@steps) {

    my @frame = grep {
      $_->is_visible($eyes);
    } map {
      _setup_plane_from_point_indices($step, $_);
    } @{$polygons};

    push @planes, \@frame;
  }

  return @planes;
}

sub _setup_plane_from_point_indices {
  my ($step, $polygon) = @_;

  my @points = @{$polygon->{vertices}};
  my @vertices = @{$step}[@points];

  my $colour = $polygon->{colour};

  my $plane = Vector3D::Plane::Easy->new(fill => $colour, vertices => \@vertices);

  return $plane;
}

sub _get_steps_as_c64_lines {
  my ($self) = @_;

  my $width  = $self->{width};
  my $height = $self->{height};

  my @planes = @{$self->{planes}};

  my @steps;

  for my $plane (@planes) {

    my @lines = map {
      my $colour = $_->{colour};
      [
        map {
          [ int ((int ($_->[0]) + $translate_x) / $divide_x), int ($_->[1]) + $translate_y + 0, $colour ]
        } @{$_->{vertices}}
      ]
    } map {
      {
        vertices => [
          map {
          translate_point($_, [ - int $width, - int $height, 0 ] )
          } @{$_->{vertices}}
        ],
        colour   => $_->{colour},
      }
    } map {
      {
        vertices => [ $_->get_vertices() ],
        colour   => $_->get_colour(),
      }
    } @{$plane};

    push @steps, \@lines;
  }

  return @steps;
}

sub _rearrange_steps_by_colours {
  my ($self, $steps) = @_;

  for my $step (@{$steps}) {

    my $colours = $self->{colours};

    my %new_step = map { $_ => [] } @{$colours};

    for my $lines (@{$step}) {

        my $colour = $lines->[0]->[2];

        my @points = map {
          [ $_->[0], $_->[1] ]
        } @{$lines};

        $new_step{$colour} = \@points;
    }

    $step = \%new_step;
  }

  return;
}

sub _step_to_lines {
  my ($self, $step, $index) = @_;

  my %lines;

  # Connect all points together and build line from all subsequent points:
  for my $colour (keys %{$step}) {

    my @points = @{$step->{$colour}};

    # Make each created shape smaller by one pixel per every utmost vertex:
    _reduce_object_size_by_1(\@points);

    my @single_colour_lines;

    for (my $i = 0; $i < @points; $i++) {

      my $point1 = $points[$i];
      my $point2 = $points[($i + 1) & (scalar (@points) - 1)];

      # Swap both points, so that @point1 always appears left to the @point2,
      # and in case they build a vertical line, then @point1 always appears
      # above the @point2:
      my @single_colour_line = map { @{$_} } sort { $a->[0] <=> $b->[0] || $a->[1] <=> $b->[1] } ($point1, $point2);

      # Push additional precalculated data into each single line item:
      my @additional_precalc_data = $self->_precalculate_additional_line_data(@single_colour_line);
      push @single_colour_line, @additional_precalc_data;

      push @single_colour_lines, \@single_colour_line;
    }

    $lines{$colour} = \@single_colour_lines;
  }

  _eliminate_single_line_items(\%lines);

  return \%lines;
}

sub _precalculate_additional_line_data {
  my ($self, $x1, $y1, $x2, $y2) = @_;

  my $deltax = $x2 - $x1;
  my $deltay = $y1 < $y2 ? $y2 - $y1 : $y1 - $y2;

  my $deltaerr = _precalculate_deltaerr($deltax, $deltay);

  my @additional_precalc_data = ($deltax, $deltay, $deltaerr);

  return @additional_precalc_data;
}

sub _precalculate_deltaerr {
  my ($deltax, $deltay) = @_;

  my ($dividend, $divisor);

  # IF ABS(Y2 - Y1) > X2 - X1 THEN deltaerr = deltax / deltay * $100
  if (abs ($deltay) > $deltax) {
    ($dividend, $divisor) = ($deltax, $deltay);
  }
  # IF ABS(Y2 - Y1) < X2 - X1 THEN deltaerr = deltay / deltax * $100
  else {
    ($dividend, $divisor) = ($deltay, $deltax);
  }

  if ($divisor == 0x00) {
    return 0x00;
  }
  else {
    return int ($dividend / $divisor * 0x0100) & 0xff;
  }
}

sub _angle_between_lines {
  my ($line1, $line2) = @_;

  my ($u2, $u1) = ( $line1->[1] - $line1->[3], $line1->[0] - $line1->[2] );
  my ($v2, $v1) = ( $line2->[3] - $line2->[1], $line2->[2] - $line2->[0] );

  my $angle_between_vecs = ($u1 * $v1 + $u2 * $v2) / (sqrt ($u1 * $u1 + $u2 * $u2) * sqrt ($v1 * $v1 + $v2 * $v2));

  return abs $angle_between_vecs;
}

sub _lines_are_ordered_from_left_to_right {
  my ($line1, $line2) = @_;

  my $line_matching_x = $line1->[2];
  my $line1_other_x   = $line1->[0];
  my $line2_other_x   = $line2->[2];

  if ($line1_other_x < $line_matching_x && $line2_other_x > $line_matching_x) {
    return 1;
  }
  elsif ($line2_other_x < $line_matching_x && $line1_other_x > $line_matching_x) {
    return 1;
  }
  else {
    return 0;
  }
}

sub _eliminate_single_line_items {
  my ($lines) = @_;

  for my $line (values %{$lines}) {

    my $uniq_x = uniq map { $_->[0] } @{$line};
    my $uniq_y = uniq map { $_->[1] } @{$line};

    $line = [] if $uniq_x < 2 || $uniq_y < 2;
  }

  return;
}

sub _dump_step_lines {
  my (%lines) = @_;

  my @colours = keys %lines;

  for (my $i = 0; $i < @colours; $i++) {

    my $colour = $colours[$i];

    printf "%d. %s\n", $i + 1, $colour;

    my $lines = $lines{$colour};

    for my $line (@{$lines}) {
      if (@{$line} == 2) {
        printf "   [ \$%02x, \$%02x ]\n", @{$line};
      }
      elsif (@{$line} == 4) {
        printf "   [ \$%02x, \$%02x ] -> [ \$%02x, \$%02x ]\n", @{$line};
      }
    }
  }

  return;
}

sub _reduce_object_size_by_1 {
  my ($points) = @_;

  return unless @{$points};

  my ($min_x, $min_y, $max_x, $max_y) = _get_min_max_point_coords($points);

  for my $point (@{$points}) {

    $point->[0] += 1 if $point->[0] == $min_x;
    $point->[0] -= 1 if $point->[0] == $max_x;

    $point->[1] += 1 if $point->[1] == $min_y;
    $point->[1] -= 1 if $point->[1] == $max_y;
  }

  return;
}

sub _get_min_max_point_coords {
  my ($points) = @_;

  # Initial conditions:
  my ($min_x, $min_y) = ($points->[0]->[0], $points->[0]->[1]);
  my ($max_x, $max_y) = ($min_x, $min_y);

  # Find min and max values amongst all points:
  for my $point (@{$points}) {

    $min_x = $point->[0] if $point->[0] < $min_x;
    $min_y = $point->[1] if $point->[1] < $min_y;

    $max_x = $point->[0] if $point->[0] > $max_x;
    $max_y = $point->[1] if $point->[1] > $max_y;
  }

  return ($min_x, $min_y, $max_x, $max_y);
}

sub _draw_lines {
  my ($self, $lines) = @_;

  my @preview = $self->_setup_empty_preview();

  my @symbols = ('@@', '%%', '##');

  for my $plane (values %{$lines}) {

    my $symbol = shift (@symbols);

    for my $line (@{$plane}) {

      _draw_line($line, $symbol, \@preview);
    }
  }

  return join "\n", map { join ' ', @{$_} } @preview;
}

sub _draw_planes {
  my ($self, @planes) = @_;

  my @preview = $self->_setup_empty_preview();

  my @symbols = ('@@', '%%', '##');

  for my $plane (@planes) {

    my $symbol = shift (@symbols);

    for my $line (@{$plane}) {

      _draw_line($line, $symbol, \@preview);
    }
  }

  return join "\n", map { join ' ', @{$_} } @preview;
}

sub _setup_empty_preview {
  my ($self) = @_;

  my ($min_x, $min_y, $max_x, $max_y) = @{ $self->check_min_max_coords([ $self->get_steps ]) }{qw(min_x min_y max_x max_y)};

  my $last_x = $min_x + $max_x;
  my $last_y = $min_y + $max_y;

  my @preview = map { [ map { '--' } ( 0 .. $last_x ) ] } ( 0 .. $last_y );

  return @preview;
}

sub _draw_line {
  my ($line, $symbol, $preview) = @_;

  my ($x1, $y1, $x2, $y2) = @{$line};

  if ($x1 > $x2) {
    ($x1, $y1, $x2, $y2) = ($x2, $y2, $x1, $y1);
  }

  my $dx = $x2 - $x1;

  if ($dx == 0) {

    $preview->[$y1]->[$x1] = $symbol;
    $preview->[$y2]->[$x2] = $symbol;
  }
  else {

    my $dy = ($y2 - $y1) / $dx;

    for (my $x = $x1; $x <= $x2; $x++) {

      my $y = $y1 + $dy * ($x - $x1);

      $preview->[$y]->[$x] = $symbol;
    }
  }

  return;
}

1;
