package Vector3D::Easy;

=head1 NAME

Vector3D::Easy - Simplified 3D object transformations and projection routines

=head1 SYNOPSIS

  # Set up a 3D point:
  my $point = [ 1, 1, 1 ];

  # Translate point by a given vector:
  my $point_translated = translate_point($point, [ 2, -1, -3 ]); # 1, 1, 1 => 3, 0, -2

  # Scale point by a given factor:
  my $point_scaled = scale_point($point, 2); # 1, 1, 1 => 2, 2, 2

  # Rotate point by a given angle around the X axis:
  my $point_rotated_x = rotate_point($point, [ $pi / 2, 0, 0 ]); # 1, 1, 1 => 1, -1, 1

  # Rotate point by a given angle around the Y axis:
  my $point_rotated_y = rotate_point($point, [ 0, $pi / 2, 0 ]); # 1, 1, 1 => 1, 1, -1

  # Rotate point by a given angle around the Z axis:
  my $point_rotated_z = rotate_point($point, [ 0, 0, $pi / 2 ]); # 1, 1, 1 => -1, 1, 1

  # Cast point from 3D space into 2D plane using perspective projection:
  my $point_casted = perspective_cast($point, $eyes);

=cut

use utf8;
use bytes;
use strict;
use warnings;

use Data::Dumper;
use List::Util qw(sum);

use base qw(Exporter);
our %EXPORT_TAGS = ();
$EXPORT_TAGS{transform} = [ qw(&translate_point &scale_point &rotate_point) ];
$EXPORT_TAGS{project} = [ qw(&perspective_cast) ];
$EXPORT_TAGS{all} = [ @{ $EXPORT_TAGS{transform} }, @{ $EXPORT_TAGS{project} } ];
our @EXPORT_OK = @{ $EXPORT_TAGS{all} };
our @EXPORT = qw();

sub scale_point {
  my ($point, $factor) = @_;

  my @scale_matrix = _get_scale_matrix($factor);

  my @point_matrix = _point_to_matrix($point);

  my $scaled_matrix = _multiply_matrices(\@scale_matrix, \@point_matrix);

  my $scaled_point = _matrix_to_point($scaled_matrix);

  return $scaled_point;
}

sub translate_point {
  my ($point, $vector) = @_;

  my @translate_matrix = _get_translate_matrix($vector);

  my @point_matrix = _point_to_matrix_extended($point);

  my $translated_matrix = _multiply_matrices(\@translate_matrix, \@point_matrix);

  my $translated_point = _matrix_extended_to_point($translated_matrix);

  return $translated_point;
}

sub rotate_point {
  my ($point, $angles) = @_;

  my @rotate_x_matrix = _get_rotate_x_matrix($angles->[0]);
  my @rotate_y_matrix = _get_rotate_y_matrix($angles->[1]);
  my @rotate_z_matrix = _get_rotate_z_matrix($angles->[2]);

  my @point_matrix = _point_to_matrix($point);

  my $rotated_x_matrix = _multiply_matrices(\@rotate_x_matrix, \@point_matrix);
  my $rotated_y_matrix = _multiply_matrices(\@rotate_y_matrix, $rotated_x_matrix);
  my $rotated_z_matrix = _multiply_matrices(\@rotate_z_matrix, $rotated_y_matrix);

  my $rotated_point = _matrix_to_point($rotated_z_matrix);

  return $rotated_point;
}

sub perspective_cast {
  my ($point, $eyes) = @_;

  my ($px, $py, $pz) = @{$point};
  my ($ex, $ey, $ez) = @{$eyes};

  my $sx = $ez * ($px - $ex) / ($ez + $pz) + $ex;
  my $sy = $ez * ($py - $ey) / ($ez + $pz) + $ey;

  return [$sx, $sy, 0];
}

sub _get_scale_matrix {
  my ($factor) = @_;

  my @scale_matrix = (
    [ $factor, 0, 0 ],
    [ 0, $factor, 0 ],
    [ 0, 0, $factor ],
  );

  return @scale_matrix;
}

sub _get_translate_matrix {
  my ($vector) = @_;

  my ($vx, $vy, $vz) = @{$vector};

  my @translate_matrix = (
    [ 1, 0, 0, $vx ],
    [ 0, 1, 0, $vy ],
    [ 0, 0, 1, $vz ],
    [ 0, 0, 0, 1   ],
  );

  return @translate_matrix;
}

sub _get_rotate_x_matrix {
  my ($angle) = @_;

  my @rotate_x_matrix = (
    [ 1, 0,              0              ],
    [ 0, + cos ($angle), - sin ($angle) ],
    [ 0, + sin ($angle), + cos ($angle) ],
  );

  return @rotate_x_matrix;
}

sub _get_rotate_y_matrix {
  my ($angle) = @_;

  my @rotate_y_matrix = (
    [ + cos ($angle), 0, + sin ($angle) ],
    [ 0,              1, 0              ],
    [ - sin ($angle), 0, + cos ($angle) ],
  );

  return @rotate_y_matrix;
}

sub _get_rotate_z_matrix {
  my ($angle) = @_;

  my @rotate_z_matrix = (
    [ + cos ($angle), - sin ($angle), 0 ],
    [ + sin ($angle), + cos ($angle), 0 ],
    [ 0,              0,              1 ],
  );

  return @rotate_z_matrix;
}

sub _multiply_matrices {
  my ($matrix1, $matrix2) = @_;

  my $rows = scalar @{$matrix1};
  my $cols = scalar @{$matrix2->[0]};

  die "Invalid matrix multiplication" unless scalar @{$matrix1->[0]} == scalar @{$matrix2};

  my $count = scalar @{$matrix2};

  my @result;

  for (my $i = 0; $i < $rows; $i++) {
    for (my $j = 0; $j < $cols; $j++) {

      my $value = sum(map { $matrix1->[$i]->[$_] * $matrix2->[$_]->[$j] } (0 .. $count - 1)) + 0;

      $result[$i]->[$j] = $value;
    }
  }

  return \@result;
}

sub _point_to_matrix {
  my ($point) = @_;

  my @matrix = map { [ $_ ] } @{$point};

  return @matrix;
}

sub _point_to_matrix_extended {
  my ($point) = @_;

  my @matrix = _point_to_matrix($point);

  push @matrix, [ 1 ];

  return @matrix;
}

sub _matrix_to_point {
  my ($matrix) = @_;

  my @point = map { $_->[0] } @{$matrix};

  return \@point;
}

sub _matrix_extended_to_point {
  my ($matrix) = @_;

  my $point = _matrix_to_point($matrix);

  pop @{$point};

  return $point;
}

1;
