package Vector3D::Plane::Easy;

=head1 NAME

Vector3D::Plane::Easy - Simplified plane object operations in a 3D space

=head1 SYNOPSIS

  # Construct new plane object:
  my @vertices = (
    [ -1, +1, -1 ],
    [ +1, +1, -1 ],
    [ +1, -1, -1 ],
    [ -1, -1, -1 ],
  );
  my $plane = Vector3D::Plane::Easy->new(fill => '#cccccc', vertices => \@vertices);

  # Verify if a plane is visible by an observer at a given position:
  my $is_visible = $plane->is_visible($eyes);

  # Get all subsequent plane vertices as a list of [x, y, z] arrays:
  my @lines = $plane->get_vertices();

  # Get configured plane fill colour:
  my $colour = $plane->get_colour();

=cut

use bytes;
use strict;
use utf8;
use warnings;

use Data::Dumper;
use List::Util qw(sum);

sub new {
  my ($this, %args) = @_;
  my $class = ref $this || $this;
  my $self = {};
  bless $self, $class;
  $self->_initialize(%args);
  return $self;
}

sub _initialize {
  my ($self, %args) = @_;

  $self->{fill} = $args{fill};
  $self->{vertices} = $args{vertices};

  return;
}

sub is_visible {
  my ($self, $eyes) = @_;

  my $vertices = $self->{vertices};

  # Take the dot-product of the view-vector (imagine an arrow coming out of the
  # camera) and the surface normal. If it's negative, then the face is visible:
  my $view_vector = _sub_vectors($vertices->[1], $eyes);
  my $normal_vector = $self->_normal_vector();

  my $dot_product = _dot_product($view_vector, $normal_vector);

  return $dot_product < 0;
}

sub get_vertices {
  my ($self) = @_;

  my $vertices = $self->{vertices};

  my @lines = @{$vertices};

  return @lines;
}

sub get_colour {
  my ($self) = @_;

  my $colour = $self->{fill};

  return $colour;
}

sub _sub_vectors {
  my ($vector1, $vector2) = @_;

  die "Invalid dot product arguments" unless scalar @{$vector1} == scalar @{$vector2};

  my $count = scalar @{$vector2};

  my @result = map { $vector1->[$_] - $vector2->[$_] } (0 .. $count - 1);

  return \@result;
}

sub _normal_vector {
  my ($self) = @_;

  my $vertices = $self->{vertices};

  my $vertex1 = $vertices->[0];
  my $vertex2 = $vertices->[1];
  my $vertex3 = $vertices->[2];

  my $u = _sub_vectors($vertex3, $vertex2);
  my $v = _sub_vectors($vertex1, $vertex2);

  my ($ux, $uy, $uz) = @{$u};
  my ($vx, $vy, $vz) = @{$v};

  # [  i   j   k  ]
  # [ $ux $uy $uz ]
  # [ $vx $vy $vz ]
  my $nx = $uy * $vz - $vy * $uz;
  my $ny = $vx * $uz - $ux * $vz;
  my $nz = $ux * $vy - $vx * $uy;
  my $n = [$nx, $ny, $nz];

  return $n;
}

sub _dot_product {
  my ($vector1, $vector2) = @_;

  die "Invalid dot product arguments" unless scalar @{$vector1} == scalar @{$vector2};

  my $count = scalar @{$vector2};

  my $value = sum(map { $vector1->[$_] * $vector2->[$_] } (0 .. $count - 1)) + 0;

  return $value;
}

1;
