#!/usr/bin/perl
#
# generate-put-pixel-lookup-tables.pl
#
#   usage: perl scripts/generate-put-pixel-lookup-tables.pl > target/put-pixel-lookup-tables.src

use bytes;
use strict;
use warnings;

use Readonly;
Readonly our $comment_line => q{;} . q{-} x 51 . qq{\n};

print $comment_line;

my ($lookup_table_x_memory, $lookup_table_x_offset_color_1, $lookup_table_x_offset_color_2, $lookup_table_x_offset_color_3) = setup_lookup_tables();
print generate_lookup_tables_dump($lookup_table_x_memory, q{CHARSET_OUTLINE_1}, q{put_pixel_1_memaddr_});
print generate_lookup_tables_dump($lookup_table_x_memory, q{CHARSET_OUTLINE_2}, q{put_pixel_2_memaddr_});
print generate_offset_tables_dump($lookup_table_x_offset_color_1, $lookup_table_x_offset_color_2, $lookup_table_x_offset_color_3);

sub setup_lookup_tables {
    my @lookup_table_x_memory = ();
    my @lookup_table_x_offset_color_1 = ();
    my @lookup_table_x_offset_color_2 = ();
    my @lookup_table_x_offset_color_3 = ();

    for (my $char = 0; $char < 8; $char++) {
        my @x_offsets = (0x40, 0x80, 0xc0);
        for (my $x = 0; $x < 4; $x++) {
            my $font_data_memory_address = $char * 0x40;
            my $font_data_memory_address_formatted = sprintf q{$%04x}, $font_data_memory_address;
            push @lookup_table_x_memory, $font_data_memory_address_formatted;
            push @lookup_table_x_offset_color_1, sprintf q{%02x}, $x_offsets[0];
            push @lookup_table_x_offset_color_2, sprintf q{%02x}, $x_offsets[1];
            push @lookup_table_x_offset_color_3, sprintf q{%02x}, $x_offsets[2];
            @x_offsets = map { $_ >> 2 } @x_offsets;
        }
    }

    return (\@lookup_table_x_memory, \@lookup_table_x_offset_color_1, \@lookup_table_x_offset_color_2, \@lookup_table_x_offset_color_3);
}

sub generate_lookup_tables_dump {
    my ($lookup_table_x_memory, $memory_base_address, $lookup_table_name_prefix) = @_;

    my @table_dump_config = (
        {
            lookup_table => $lookup_table_x_memory,
            table_name   => sprintf (q{%sx_lo}, $lookup_table_name_prefix),
            base_address => sprintf (q{<(%s + }, $memory_base_address),
            values_dump  => q{SINGLE},
            additional_indentation => 0,
        },
        {
            lookup_table => $lookup_table_x_memory,
            table_name   => sprintf (q{%sx_hi}, $lookup_table_name_prefix),
            base_address => sprintf (q{>(%s + }, $memory_base_address),
            values_dump  => q{SINGLE},
            additional_indentation => 0,
        },
    );

    my $lookup_tables_dump = dump_source_code(@table_dump_config);

    return $lookup_tables_dump;
}

sub generate_offset_tables_dump {
    my ($lookup_table_x_offset_color_1, $lookup_table_x_offset_color_2, $lookup_table_x_offset_color_3) = @_;

    my @table_dump_config = (
        {
            lookup_table => $lookup_table_x_offset_color_1,
            table_name   => sprintf (q{put_pixel_colour_1_bits}),
            base_address => q{},
            values_dump  => q{MULTIPLE},
            additional_indentation => 3,
        },
        {
            lookup_table => $lookup_table_x_offset_color_2,
            table_name   => sprintf (q{put_pixel_colour_2_bits}),
            base_address => q{},
            values_dump  => q{MULTIPLE},
            additional_indentation => 3,
        },
        {
            lookup_table => $lookup_table_x_offset_color_3,
            table_name   => sprintf (q{put_pixel_colour_3_bits}),
            base_address => q{},
            values_dump  => q{MULTIPLE},
            additional_indentation => 3,
        },
    );

    my $lookup_tables_dump = dump_source_code(@table_dump_config);

    return $lookup_tables_dump;
}

sub dump_source_code {
    my (@table_dump_config) = @_;

    my $lookup_tables_dump = '';

    for my $config (@table_dump_config) {
        my $lookup_table           = $config->{lookup_table};
        my $table_name             = $config->{table_name};
        my $base_address           = $config->{base_address};
        my $values_dump            = $config->{values_dump};
        my $additional_indentation = $config->{additional_indentation};

        $lookup_tables_dump .= $table_name . q{ };

        if ($values_dump eq q{SINGLE}) {
            for (my $x = 0; $x < @{$lookup_table}; $x++) {
                my $value = sprintf q{%s%s)}, $base_address, $lookup_table->[$x];
                unless ($x == 0) {
                    $lookup_tables_dump .= q{ } x 25;
                    $lookup_tables_dump .= q{ } x $additional_indentation;
                }
                $lookup_tables_dump .= q{.db } . $value . qq{\n};
            }
        } elsif ($values_dump eq q{MULTIPLE}) {
            my $bytes_per_line = 6;
            for (my $x = 0; $x < @{$lookup_table}; $x = $x + $bytes_per_line) {
                my @line;
                for (my $i = 0; $i < $bytes_per_line; $i++) {
                    next unless defined $lookup_table->[$x + $i];
                    my $value = sprintf q{%s$%s}, $base_address, $lookup_table->[$x + $i];
                    push @line, $value;
                }
                unless ($x == 0) {
                    $lookup_tables_dump .= q{ } x 21;
                    $lookup_tables_dump .= q{ } x $additional_indentation;
                }
                $lookup_tables_dump .= q{ .db } . join (q{,}, @line) . qq{\n};
            }
        }
        else {
            die sprintf q{Invalid values dump requested: "%s"}, $values_dump;
        }

        $lookup_tables_dump .= $comment_line;
    }

    return $lookup_tables_dump;
}

1;

__END__