#!/usr/bin/perl
#
# Usage:
#
#   $ scripts/setup-compressed-images.pl <INPUT_FILE_PATTERN> <OUTPUT_FILE_PATTERN> <INCLUDE_SRC_FILE>
#
# Example:
#
#   $ scripts/setup-compressed-images.pl images/olimp#NUM#.prg target/olimp#NUM#.prg target/compressed-images.inc

use utf8;
use bytes;
use strict;
use warnings;

local $| = 1;

use lib q{../../tools/lib};

use Archive::ByteBoozer qw(crunch);
use Data::Dumper;
use IO::File;
use Misc::Util qw(modify_c64_file_loading_address);
use Readonly;

our @BITMAP_ADDRESSES = (0x6000, 0xa000);
our @FREE_MEMORY = ([0x3b00, 0x5000], [0x8400, 0xa000], [0xc000, 0xd000]);

my ($input_pattern, $output_pattern, $include_file) = @ARGV;

(my $glob_input_pattern = $input_pattern) =~ s/#NUM#/*/;
my @input_files = sort glob qq{$glob_input_pattern};

for my $input_file (@input_files) {

    (my $index_input_pattern = $input_pattern) =~ s/#NUM#/(.*)/;
    my ($index) = ($input_file =~ m/$index_input_pattern/);

    (my $output_file = $output_pattern) =~ s/#NUM#/$index/;

    print qq{Compressing "$input_file" to "$output_file"...};

    compress_picture($input_file, $output_file, $include_file, $index);

    print qq{ OK!\n};
}

sub compress_picture {
    my ($source, $target, $include_file, $index) = @_;

    # Setup temporary file with the content of a source file and screen colours:
    my ($temp_file, $data) = setup_temporary_file($source, $target);

    my $original = new IO::File $temp_file, "r";
    my $crunched = new IO::File $target, "w";
    crunch(source => $original, target => $crunched);

    my $fh = new IO::File $target, 'r';
    die "Failed to read file: $target" unless defined $fh;

    my $compressed_bytes;
    $fh->binmode(':bytes');
    sysread $fh, $compressed_bytes, 0x0002;
    sysread $fh, $compressed_bytes, 0x1f40;
    $fh->close;

    my $compressed_length = length $compressed_bytes;

    my $start_address = update_free_memory($compressed_length);
    modify_c64_file_loading_address($target, $start_address);

    write_include_file($include_file, $start_address, $target, $index);

    return;
}

sub write_include_file {
    my ($filename, $start_address, $binclude_filename, $index) = @_;

    my $text = ";---------------------------------------\n";
    $text .= sprintf "                *= \$%04x\n", $start_address;
    $text .= ";---------------------------------------\n";
    $text .= sprintf "compressed_picture_%02d\n", $index;
    $text .= ".binclude \"$binclude_filename\",2\n";
    $text .= ";---------------------------------------\n\n";

    open my $fh, '>>:utf8', $filename or die "Cannot open file \"$filename\" for writing!\n";
    print $fh $text;
    close $fh or die "Cannot close file \"$filename\" opened for writing!\n";

    return;
}

sub update_free_memory {
    my ($compressed_length) = @_;

    my $start_address = $FREE_MEMORY[0]->[0];
    my $end_address = $FREE_MEMORY[0]->[1];

    my $new_start_address = $start_address + $compressed_length;

    if ($new_start_address < $end_address) {
        $FREE_MEMORY[0]->[0] = $new_start_address;
        return $start_address;
    }
    else {
        shift @FREE_MEMORY;
        return update_free_memory($compressed_length);
    }
}

sub setup_temporary_file {
    my ($source, $target) = @_;

    my $fh = new IO::File $source, 'r';
    die "Failed to read file: $source" unless defined $fh;

    my $bitmap_bytes;
    $fh->binmode(':bytes');
    sysread $fh, $bitmap_bytes, 0x0002;
    sysread $fh, $bitmap_bytes, 0x1f40;
    $fh->close;

    my $bitmap_address = shift @BITMAP_ADDRESSES;
    push @BITMAP_ADDRESSES, $bitmap_address;

    my $final_bytes = join '', map { chr 0x00 } (0x0000 .. 0x1f3f);
    substr ($final_bytes, 0x0000, 0x1f40) = $bitmap_bytes;
    $final_bytes = address_to_bytes($bitmap_address) . $final_bytes;

    my $temp_file = $target . '.tmp';

    open $fh, '>:bytes', $temp_file or die "Cannot open file \"$temp_file\" for writing!\n";
    binmode $fh;
    syswrite $fh, $final_bytes, length $final_bytes;
    close $fh or die "Cannot close file \"$temp_file\" opened for writing!\n";

    return $temp_file;
}

sub address_to_bytes {
    my ($addr) = @_;
    my $lo = $addr & 0x00ff;
    my $hi = ($addr >> 0x08) & 0x00ff;
    my @bytes = map { chr } ($lo, $hi);
    my $bytes = join '', @bytes;
    return $bytes;
}
