package Demo::Config;

use utf8;
use bytes;
use strict;
use warnings;

# use base qw(Exporter);
# our %EXPORT_TAGS = ();
# $EXPORT_TAGS{binary} = [ qw(&convert_binary_data_to_hex_values) ];
# $EXPORT_TAGS{all} = [ @{ $EXPORT_TAGS{binary} } ];
# our @EXPORT_OK = @{ $EXPORT_TAGS{all} };
# our @EXPORT = qw();

use Data::Dumper;
use FindBin qw($Bin);
use Readonly;
use YAML::Syck;

Readonly our $CONFIG_FILE => "$Bin/../../src/config.yml";

=head2 new

Build new demo configuration object:

  my $config = Demo::Config->new();

=cut

sub new {
  my ($this, %args) = @_;
  my $class = ref $this || $this;
  my $self = {};
  bless $self, $class;
  $self->_initialize(%args);
  return $self;
}

sub _initialize {
  my ($self, %args) = @_;
  my $config_file = $args{config_file} || $CONFIG_FILE;
  my $config = LoadFile($config_file)->{demo};
  $self->{config} = $config;
  return;
}

=head2 loading_address

Fetch loading address for a given part name:

  my $part_name = '02-full-scroll-d016';
  my $loading_address = $config->loading_address($part_name);

=cut

sub loading_address {
  my ($self, $part_name) = @_;
  my $part_config = $self->_part_config($part_name);
  my $loading_address = $part_config->{$part_name}->{load_addr};
  $loading_address =~ s/^\$//;
  return $loading_address;
}

=head2 setup_part_file_names

Setup loading file names for all configured demo parts:

  my %part_names = $config->setup_part_file_names();

=cut

sub setup_part_file_names {
  my ($self) = @_;
  my %part_names;
  my $config = $self->{config};
  my @part_names = $self->part_names();
  my $file_index = 1;
  while (@part_names) {
    my $file_name = unpack 'H2', chr $file_index++;
    my $part_name = splice @part_names, 0, 1;
    $part_names{$part_name} = $file_name;
  }
  return %part_names;
}

=head2 part_file_names

Fetch file names for all configured demo parts:

  my @file_names = $config->part_file_names();

You can also get file names as a source code text:

  my $file_names = $config->part_file_names(as_source_code => 1);

=cut

sub part_file_names {
  my ($self, %args) = @_;
  my $as_source_code = $args{as_source_code};
  my (@file_names, @first_letters, @second_letters);
  my %part_names = $self->setup_part_file_names();
  my @part_names = $self->part_names();
  while (my $part_name = shift @part_names) {
    my $file_name = $part_names{$part_name};
    push @file_names, {
      value   => $file_name,
      comment => $part_name,
    };
    my ($first_letter, $second_letter) = map { sprintf '"%s"', $_ } ($file_name =~ m/^(\w)(\w)$/);
    push @first_letters, {
      value   => $first_letter,
      comment => $part_name,
    };
    push @second_letters, {
      value   => $second_letter,
      comment => $part_name,
    };
  }
  if ($as_source_code) {
    my $first_letters = $self->_to_source_code(
      data        => \@first_letters,
      label       => 'part_loading_first_letters',
      lo_hi       => 0,
      dump_opcode => '.dp',
      skip_last   => 1,
    );
    my $second_letters = $self->_to_source_code(
      data        => \@second_letters,
      label       => 'part_loading_second_letters',
      lo_hi       => 0,
      dump_opcode => '.dp',
      skip_last   => 0,
    );
    return $first_letters . $second_letters;
  }
  else {
    return @file_names;
  }
}

=head2 part_loading_addresses

Fetch loading addresses for all configured demo parts:

  my @loading_addresses = $config->part_loading_addresses();

You can also get loading addresses as a source code text:

  my $loading_addresses = $config->part_loading_addresses(as_source_code => 1);

=cut

sub part_loading_addresses {
  my ($self, %args) = @_;
  my $as_source_code = $args{as_source_code};
  my @part_addresses;
  my $config = $self->{config};
  my @part_names = $self->part_names();
  while (my $part_name = shift @part_names) {
    my $part_config = $self->_part_config($part_name);
    my $loading_address = $part_config->{$part_name}->{load_addr};
    push @part_addresses, {
      value   => $loading_address,
      comment => $part_name,
    };
  }
  if ($as_source_code) {
    return $self->_to_source_code(data => \@part_addresses, label => 'part_loading_addresses', lo_hi => 1);
  }
  else {
    return @part_addresses;
  }
}

=head2 part_start_addresses

Fetch start addresses for all configured demo parts:

  my @start_addresses = $config->part_start_addresses();

You can also get start addresses as a source code text:

  my $start_addresses = $config->part_start_addresses(as_source_code => 1);

=cut

sub part_start_addresses {
  my ($self, %args) = @_;
  my $as_source_code = $args{as_source_code};
  my @part_addresses;
  my $config = $self->{config};
  my @part_names = $self->part_names();
  while (my $part_name = shift @part_names) {
    (my $part_name_inc = $part_name) =~ s/^(?:\d+-)?(.+)$/$1.inc/;
    (my $part_name_uc = $part_name) =~ s/^(?:\d+-)?(.+)$/uc "$1_START_ADDRESS"/e;
    $part_name_uc =~ tr/-/_/;
    my $start_address = `cat $Bin/../../src/$part_name/$part_name_inc | grep $part_name_uc | awk -F'=' '{ print \$2 }' | sed -e 's/\\s//g'`;
    chomp $start_address;
    push @part_addresses, {
      value   => $start_address,
      comment => $part_name,
    };
  }
  if ($as_source_code) {
    return $self->_to_source_code(data => \@part_addresses, label => 'part_start_addresses', lo_hi => 1);
  }
  else {
    return @part_addresses;
  }
}

=head2 part_synchro_timers

Fetch synchronisation timers for all configured demo parts:

  my @synchro_timers = $config->part_synchro_timers();

You can also get synchronisation timers as a source code text:

  my $synchro_timers = $config->part_synchro_timers(as_source_code => 1);

=cut

sub part_synchro_timers {
  my ($self, %args) = @_;
  my $as_source_code = $args{as_source_code};
  my @synchro_timers;
  my $config = $self->{config};
  my @part_names = $self->part_names();
  while (my $part_name = shift @part_names) {
    my $part_config = $self->_part_config($part_name);
    my $synchro_timer = $part_config->{$part_name}->{sync_time};
    push @synchro_timers, {
      value   => $synchro_timer,
      comment => $part_name,
    };
  }
  if ($as_source_code) {
    return $self->_to_source_code(data => \@synchro_timers, label => 'part_synchro_timers', lo_hi => 1);
  }
  else {
    return @synchro_timers;
  }
}

=head2 part_names

Fetch the list of all configured part names:

  my @part_names = $self->part_names();

Part names are basically all names of source code directories that were configured in the "config.yml" file.

=cut

sub part_names {
  my ($self) = @_;
  my $config = $self->{config};
  my @part_names = sort map { keys %$_ } @{$config};
  return @part_names;
}

sub _part_config {
  my ($self, $part_name) = @_;
  my $config = $self->{config};
  my ($part_config) = grep { exists $_->{$part_name}; } @{$config};
  return $part_config;
}

sub _to_source_code {
  my ($self, %args) = @_;
  my $data = $args{data};
  my $dump_opcode = $args{dump_opcode};
  my $lo_hi = $args{lo_hi};
  my $label = $args{label};
  my $skip_last = $args{skip_last} || 0;
  my @rows;
  if ($lo_hi) {
    my @lo_rows = $self->_to_source_code_single(before => '<', data => $data, dump_opcode => $dump_opcode, label => $label . '_lo');
    @lo_rows = $self->_add_comments_around(source_rows => \@lo_rows, skip_last => 1);
    my @hi_rows = $self->_to_source_code_single(before => '>', data => $data, dump_opcode => $dump_opcode, label => $label . '_hi');
    @hi_rows = $self->_add_comments_around(source_rows => \@hi_rows, skip_last => $skip_last);
    @rows = (@lo_rows, @hi_rows);
  }
  else {
    @rows = $self->_to_source_code_single(before => '', data => $data, dump_opcode => $dump_opcode, label => $label);
    @rows = $self->_add_comments_around(source_rows => \@rows, skip_last => $skip_last);
  }
  push @rows, "";
  my $source_code = join "\n", @rows;
  return $source_code;
}

sub _to_source_code_single {
  my ($self, %args) = @_;
  my $before = $args{before};
  my @data = @{$args{data}};
  my $dump_opcode = $args{dump_opcode} || '.db';
  my $label = $args{label};
  my @source_rows = ($label);
  for my $item (@data) {
    my $source_row = sprintf '%s%s %s%s ; %s', ' ' x 16, $dump_opcode, $before, $item->{value}, $item->{comment};
    push @source_rows, $source_row;
  }
  return @source_rows;
}

sub _add_comments_around {
  my ($self, %args) = @_;
  my @source_rows = @{$args{source_rows}};
  my $skip_last = $args{skip_last};
  my ($maxlen) = sort { $b <=> $a } map { length } @source_rows;
  my $comment = ';' . '-' x $maxlen;
  unshift @source_rows, $comment;
  splice @source_rows, 2, 0, $comment;
  unless ($skip_last) {
    push @source_rows, $comment;
  }
  return @source_rows;
}

1;

__END__