package Graphics::Builder;

=head1 NAME

Graphics::Builder - Builder class to create miscellaneous graphic images

=head1 SYNOPSIS

  # Create a new image builder object:
  my $image = Graphics::Builder->new();

  # Build image builder object from a PNG file:
  my $image = Graphics::Builder->new_from(png_file => $png_file);

  # Build image builder object from an Image::Magick object:
  my $image = Graphics::Builder->new_from(magick_object => $magick_object);

  # Write large text in the middle of an image:
  $image->write_text(size => 112, text => "TEST", top_margin => 32);

  # Cut out fragment of an image, creating a new smaller image object in return:
  my $new_image = $image->cut_out_fragment(x => 6, y => 0, width => 7, height => 10);

  # Get 8000 raw hires bitmap data bytes without screen colours (ready to save to a disk file):
  my $bytes = $image->to_hires_bitmap();

  # Convert picture to multicolour sprites data:
  my $bytes = $image->to_multicolour_sprites(rows => 3, cols => 2);

  # Get direct access to an underlying Image::Magick object:
  my $image_magick = $image->magick();

  # Preview currently generated image:
  $image->preview();

=cut

use utf8;
use bytes;
use strict;
use warnings;

use Data::Dumper;
use Image::Magick;

my %fonts = (
  'DejaVuSans-Bold'           => '/usr/share/fonts/dejavu/DejaVuSans-Bold.ttf',
  'DejaVuSansCondensed-Bold'  => '/usr/share/fonts/dejavu/DejaVuSansCondensed-Bold.ttf',
  'DejaVuSansMono-Bold'       => '/usr/share/fonts/dejavu/DejaVuSansMono-Bold.ttf',
  'FreeSans'                  => '/usr/share/fonts/gnu-free/FreeSans.ttf',
  'LiberationSans-Bold'       => '/usr/share/fonts/liberation/LiberationSans-Bold.ttf',
  'LiberationSansNarrow-Bold' => '/usr/share/fonts/liberation/LiberationSansNarrow-Bold.ttf',
);

sub new {
  my ($this, %args) = @_;
  my $class = ref $this || $this;
  my $self = {};
  bless $self, $class;
  $self->_initialize(%args);
  return $self;
}

sub _initialize {
  my ($self, %args) = @_;
  my $size = $args{size} || '320x200';
  my $image = Image::Magick->new(size => $size);
  $image->Set(antialias => 0);
  $image->ReadImage('xc:#000000');
  $self->{image} = $image;
  return;
}

sub new_from {
  my ($this, %args) = @_;
  my $self = $this->new(%args);

  my $magick_object = $args{magick_object};

  my $png_file = $args{png_file};
  my $scale_factor = $args{scale_image_by_factor};

  if ($png_file) {
    die "File does not exists: $args{png_file}" unless -e $args{png_file};
    my $image = $self->{image};
    $image->Read($png_file);
    undef $image->[0];
    if ($scale_factor) {
      $self->scale_image(by_factor => $scale_factor);
    }
    $self->_resize_to_320x200();
  }

  elsif ($magick_object) {
    $self->{image} = $magick_object;
  }

  return $self;
}

sub _resize_to_320x200 {
  my ($self) = @_;
  my $image = $self->{image};
  # Get the most used colour written into picture and use it as a background colour when extending the image size:
  my $histogram = $self->_get_histogram();
  my ($most_used_colour) = sort { $histogram->{$b} <=> $histogram->{$a} } keys %{$histogram};
  # Crop an image:
  $image->Crop(geometry => '320x200', gravity => 'NorthWest');
  # Set the image size:
  $image->Extent(geometry => '320x200', gravity => 'NorthWest', background => $most_used_colour);
  return;
}

sub magick {
  my ($self) = @_;
  my $image = $self->{image};
  return $image;
}

sub preview {
  my ($self) = @_;
  my $image = $self->{image};
  $image->Display();
  return;
}

sub scale_image {
  my ($self, %args) = @_;
  my $by_factor = $args{by_factor};
  my $image = $self->{image};
  my $width = $image->Get('width') * $by_factor;
  my $height = $image->Get('height') * $by_factor;
  $image->Scale(geometry => "${width}x${height}");
  return;
}

sub write_text {
  my ($self, %args) = @_;

  my $gravity    = $args{gravity}    || 'North';
  my $colour     = $args{colour}     || '#ffffff';
  my $font       = $args{font}       || 'DejaVuSans-Bold';
  my $size       = $args{size}       || 36;
  my $text       = $args{text}       || '';
  my $top_margin = $args{top_margin} || 0;

  my $ttf_font = exists $fonts{$font} ? $fonts{$font} : $font;

  my $image = $self->{image};
  $image->Annotate(
    text      => $text,
    geometry  => "+0+$top_margin",
    font      => $ttf_font,
    fill      => $colour,
    gravity   => $gravity,
    pointsize => $size,
  );

  $self->{last_write_colour} = $colour;

  return;
}

=head2 cut_out_fragment

Cut out fragment of an image, creating a new smaller image object in return:

  my $new_image = $image->cut_out_fragment(x => 6, y => 0, width => 7, height => 10);

Please note that all arguments are given in chars, not in pixels!

=cut

sub cut_out_fragment {
  my ($self, %args) = @_;

  my $x      = $args{x} * 0x08;
  my $y      = $args{y} * 0x08;
  my $width  = $args{width} * 0x08;
  my $height = $args{height} * 0x08;

  my $new_image = $self->{image}->Clone();
  $new_image->Crop('width' => $width, 'height' => $height, 'x' => $x, 'y' => $y, 'gravity' => 'NorthWest');

  my $new_picture = Graphics::Builder->new_from(magick_object => $new_image);
  return $new_picture;
}

sub _get_histogram {
  my ($self, %args) = @_;

  my $image = $self->{image};

  my @data = $image->Histogram();

  my %histogram;
  while (@data) {
    my ($r, $g, $b, $a, $count) = splice @data, 0, 5;
    my $colour = sprintf ('#%02x%02x%02x', map { int ($_ / 256) } ($r, $g, $b));
    $histogram{$colour} = $count;
  }

  return \%histogram;
}

sub _get_all_pixels {
  my ($self) = @_;

  my $image = $self->{image};

  my @data = $image->GetPixels(geometry => '320x200', normalize => 1, map => 'RGBA');

  my @pixels;
  while (@data) {
    my ($r, $g, $b, $a) = splice @data, 0, 4;
    my $pixel = sprintf ('#%02x%02x%02x', map { int ($_ * 255) } ($r, $g, $b));
    push @pixels, $pixel;
  }

  return \@pixels;
}

=head2 to_hires_bitmap

Get 8000 raw hires bitmap data bytes without screen colours (ready to save to a disk file):

  my $bytes = $image->to_hires_bitmap();

Additional options are supported to give more flexibility in converting images to raw bitmaps.

Do not fetch initial empty lines, instead shift an entire image data up to the top margin:

  my $bytes = $image->to_hires_bitmap(shifted_to => 'top');

=cut

sub to_hires_bitmap {
  my ($self, %args) = @_;

  my $shifted_to_top = defined $args{shifted_to} && $args{shifted_to} eq 'top';

  my $image = $self->{image};

  my $width = $image->Get('width');

  # Get the least used colour written into picture as a fallback:
  my $histogram = $self->_get_histogram();
  my ($least_used_colour) = sort { $histogram->{$a} <=> $histogram->{$b} } keys %{$histogram};

  # Get the only one last colour that was used to write something into picture:
  my $colour = $self->{last_write_colour} || $least_used_colour;

  my $pixels = $self->_get_all_pixels();

  # If user requested shifting an entire image data up to the top margin, first
  # let's remove all empty pixel lines from the top of a picture:
  $self->_remove_empty_pixel_lines_from_top($pixels, $colour) if $shifted_to_top;

  my $bytes;
  for (my $y = 0; $y < 0x19; $y++) {
    for (my $x = 0; $x < 0x28; $x++) {
      for (my $j = 0; $j < 0x08; $j++) {
        my $byte = 0x00;
        for (my $i = 0; $i < 0x08; $i++) {
          my $cx = $x * 0x08 + $i;
          my $cy = $y * 0x08 + $j;
          # my @pixels = $image->GetPixel('x' => $cx, 'y' => $cy);
          # my $pixel = sprintf ('#%02x%02x%02x', map { int ($_ * 255) } @pixels);
          my $pixel = $pixels->[$cx + $cy * $width] || '';

          $byte <<= 0x01;
          if ($pixel eq $colour) {
            $byte |= 0x01;
          }
        }
        $bytes .= chr $byte;
      }
    }
  }

  return $bytes;
}

sub _remove_empty_pixel_lines_from_top {
  my ($self, $pixels, $colour) = @_;

  REMOVE:
  while (@{$pixels}) {
    for (my $x = 0; $x < 0x0140; $x++) {
      my $pixel = $pixels->[$x];
      if ($pixel eq $colour) {
        last REMOVE;
      }
    }
    splice @{$pixels}, 0x0000, 0x0140;
  }

  return;
}

=head2 to_multicolour_sprites

Convert picture to multicolour sprites data:

  my $bytes = $image->to_multicolour_sprites(rows => 3, columns => 2, colours => [0x00, 0x0b, 0x0c]);

=cut

sub to_multicolour_sprites {
  my ($self, %args) = @_;

  my $rows = $args{rows};
  my $cols = $args{cols};

  die "Not implemented yet!";
}

sub _setup_colour_map {
}

1;
