package Graphics::Factory;

=head1 NAME

Graphics::Factory - Factory class to create miscellaneous graphic images

=head1 SYNOPSIS

  # Fetch data bytes required to create an empty hires picture:
  my $bytes = Graphics::Factory->empty_hires_bytes();

  # Create an empty hires picture:
  my $picture = Graphics::Factory->empty_hires();

  # Build a hires picture out of a partial data and given dimensions:
  my $picture = Graphics::Factory->hires(data => $bytes, dimensions => '40x10');

=cut

use utf8;
use bytes;
use strict;
use warnings;

use Data::Dumper;
use Graphics::HiRes;

sub empty_hires {
  my ($pkg) = @_;
  my $bytes = $pkg->empty_hires_bytes();
  my $picture = Graphics::HiRes->new(data => $bytes);
  return $picture;
}

sub empty_hires_bytes {
  my ($pkg) = @_;
  my $bitmap_bytes = join '', map { chr 0x00 } ( 0x00 .. 0x1f3f );
  my $screen_bytes = join '', map { chr 0xbc } ( 0x00 .. 0x03e7 );
  my $bytes = $bitmap_bytes . $screen_bytes;
  return $bytes;
}

sub hires {
  my ($pkg, %arg) = @_;

  my $data = $arg{data};
  my ($orig_width, $orig_height) = $arg{dimensions} =~ m/^(\d+)x(\d+)$/;
  my $bytes = $pkg->empty_hires_bytes();

  my $height = 0;
  while ($height < $orig_height) {
    substr ($bytes, $height * 0x0140, $orig_width * 0x08) = substr $data, $height * $orig_width * 0x08, $orig_width * 0x08;
    $height++;
  }

  my $picture = Graphics::HiRes->new(data => $bytes);
  return $picture;
}

1;

__END__