package Graphics::HiRes;

=head1 NAME

Graphics::HiRes - A base class for any hires image operations

=head1 SYNOPSIS

  # Fetch row (0..24) raw (only bitmap pixels) data:
  my $data = $picture->get_raw_data(row => 0);

  # Fetch row (0..24) raw (only bitmap pixels) data limited by given columns (0..39):
  my $data = $picture->get_raw_data(row => 0, cols => [ 10 .. 29 ]);

  # Fetch multiple row (0..39) raw (only bitmap pixels) data:
  my $data = $picture->get_raw_data(rows => [ 0 .. 9 ]);

  # Fetch multiple row (0..39) raw (only bitmap pixels) data limited by given columns (0..39):
  my $data = $picture->get_raw_data(rows => [ 0 .. 9 ], cols => [ 10 .. 29 ]);

  # Fetch all rows of non-empty raw (only bitmap pixels) data:
  my $data = $picture->get_non_empty_raw_data();

  # Preview graphics image:
  $picture->preview();

  # Save preview image:
  $picture->save_preview(to_file => 'image.png');

=cut

use bytes;
use strict;
use utf8;
use warnings;

use parent qw(Graphics::BaseImage);

use Data::Dumper;
use Graphics::Palette qw(get_rgb_color);
use Image::Magick;

sub new {
  my ($this, %args) = @_;
  my $class = ref $this || $this;
  my $self = {};
  bless $self, $class;
  $self->_initialize(%args);
  return $self;
}

sub _initialize {
  my ($self, %args) = @_;
  my $data = $self->_fetch_data(%args);
  $self->{rawdata} = $data;
  my $bitmap = $self->fetch_bitmap_data(rawdata => $data);
  my $screen = $self->fetch_screen_data(rawdata => $data);
  $self->_set_bitmap_data($bitmap);
  $self->_set_screen_data($screen);
  return;
}

sub fetch_bitmap_data {
  my ($self, %args) = @_;
  my $rawdata = $args{rawdata};
  my $data = substr $rawdata, 0x0000, 0x1f40;
  return $data;
}

sub fetch_screen_data {
  my ($self, %args) = @_;
  my $rawdata = $args{rawdata};
  if (length $rawdata == 0x1f40) {
    my $data = join '', map { chr 0xbc } (0x0000 .. 0x03e7);
    return $data;
  }
  else {
    my $data = substr $rawdata, 0x1f40, 0x03e8;
    return $data;
  }
}

sub get_non_empty_raw_data {
  my ($self) = @_;

  my ($from, $to) = $self->_identify_non_empty_row_range();

  my $data = $self->get_raw_data(rows => [ $from .. $to ]);

  return $data;
}

sub get_raw_data {
  my ($self, %args) = @_;

  my $cols = $args{cols};
  my $row  = $args{row};
  my $rows = $args{rows};

  my @cols = defined $cols ? @{$cols} : ( 0x00 .. 0x27 );
  my @rows = defined $rows ? @{$rows} : $row;

  my $data = $self->_get_bitmap_data();

  my $raw_data;
  for my $row (@rows) {
    for my $col (@cols) {
      my $offset = $row * 0x0140 + $col * 0x08;
      my $row_data = substr $data, 0x0140 * $row + 0x08 * $col, 0x08;
      $raw_data .= $row_data;
    }
  }

  return $raw_data;
}



sub _identify_non_empty_row_range {
  my ($self) = @_;

  my $empty_byte = chr 0x00;
  my $non_empty_line = qr/^${empty_byte}{320}$/;

  # Start at the top row and find the first non-empty line:
  my $first;
  for (my $i = 0; $i < 25; $i++) {
    my $line_data = $self->get_raw_data(row => $i);
    if ($line_data !~ $non_empty_line) {
      $first = $i;
      last;
    }
  }

  # Start at the bottom row and find the last non-empty line:
  my $last;
  for (my $i = 24; $i >= 0; $i--) {
    my $line_data = $self->get_raw_data(row => $i);
    if ($line_data !~ $non_empty_line) {
      $last = $i;
      last;
    }
  }

  return ($first, $last);
}

sub _build_preview {
  my ($self, %args) = @_;
  my $scale = $args{scale} || 2;
  my $size = '320x200';
  my $image = Image::Magick->new(size => $size);
  $image->Set(antialias => 0);
  $image->ReadImage('xc:#000000');
  my $bitmap = $self->_get_bitmap_data();
  my $screen = $self->_get_screen_data();

  for (my $x = 0; $x < 320; $x++) {
    for (my $y = 0; $y < 200; $y++) {
      my $bitmap_offset = ($y >> 0x03) * 0x0140 + ($y & 0x07) + ($x >> 0x03) * 0x08;
      my $screen_offset = ($y >> 0x03) * 0x28 + ($x >> 0x03);
      my $bitmap_bitmask = 0x80 >> ($x & 0x07);

      my $screen_byte = substr $screen, $screen_offset, 0x01;
      my $colour = (ord ($screen_byte) & 0xf0) >> 0x04;
      my $bckgrd = (ord ($screen_byte) & 0x0f);

      my $bitmap_byte = substr $bitmap, $bitmap_offset, 0x01;
      my $bit = ord ($bitmap_byte) & $bitmap_bitmask;
      my $pixel_colour = $bit ? $colour : $bckgrd;

      my $rgb_color = get_rgb_color(C64 => $pixel_colour);
      my $set_pixel = sprintf ("#%02x%02x%02x", map { $rgb_color->{$_} } qw(RED GREEN BLUE));

      $image->Set("pixel[$x,$y]" => $set_pixel);
    }
  }

  my $width = $image->Get('width') * $scale;
  my $height = $image->Get('height') * $scale;
  $image->Scale(geometry => "${width}x${height}");
  return $image;
}

sub preview {
  my ($self, %args) = @_;
  my $image = $self->_build_preview(%args);
  $image->Display();
  return;
}

sub save_preview {
  my ($self, %args) = @_;
  my $filename = $args{to_file};
  my $image = $self->_build_preview(%args);
  $image->Write($filename);
  return;
}

1;

__END__