package Graphics::MultiColour;

=head1 NAME

Graphics::MultiColour - A base class for any multicolour image operations

=head1 SYNOPSIS

  # Fetch char column (0..39) data:
  my $data = $picture->get_column_data(column => 0);

  # Fetch multiple char columns (0..39) data:
  my $data = $picture->get_columns_data(from => 0, to => 19);

=cut

use bytes;
use strict;
use utf8;
use warnings;

use parent qw(Graphics::BaseImage);

use Data::Dumper;

sub new {
  my ($this, %args) = @_;
  my $class = ref $this || $this;
  my $self = {};
  bless $self, $class;
  $self->_initialize(%args);
  return $self;
}

sub _initialize {
  my ($self, %args) = @_;
  my $file = $args{file};
  my $data = $self->_fetch_data(%args);
  $self->{rawdata} = $data;
  my $bitmap = $self->fetch_bitmap_data(rawdata => $data);
  my $screen = $self->fetch_screen_data(rawdata => $data);
  my $colors = $self->fetch_colors_data(rawdata => $data);
  my $bckgrd = $self->fetch_bckgrd_data(rawdata => $data);
  my $border = $self->fetch_border_data(rawdata => $data);
  $self->_set_bitmap_data($bitmap);
  $self->_set_screen_data($screen);
  $self->_set_colors_data($colors);
  $self->_set_bckgrd_data($bckgrd);
  $self->_set_border_data($border);
  return;
}

sub get_columns_data {
  my ($self, %args) = @_;
  my $from = $args{from};
  my $to = $args{to};
  my %columns_data;
  for my $column ($from .. $to) {
    my $column_data = $self->get_column_data(column => $column);
    while (my ($component, $data) = each %{$column_data}) {
      $columns_data{$component} .= $data;
    }
  }
  return \%columns_data;
}

sub get_column_data {
  my ($self, %args) = @_;
  my $column = $args{column};
  my %config = (
    bitmap => {
      data    => $self->_get_bitmap_data(),
      rowsize => 0x0140,
      colsize => 0x08,
    },
    screen => {
      data    => $self->_get_screen_data(),
      rowsize => 0x0028,
      colsize => 0x01,
    },
    colors => {
      data    => $self->_get_colors_data(),
      rowsize => 0x0028,
      colsize => 0x01,
    },
  );
  my $column_data = $self->_get_column_data(\%config, $column);
  return $column_data;
}

sub _get_column_data {
  my ($self, $config, $column) = @_;
  my %column_data;
  while (my ($component, $parameters) = each %{$config}) {
    my $data = $parameters->{data};
    my $rowsize = $parameters->{rowsize};
    my $colsize = $parameters->{colsize};
    my $column_data;
    for my $row (0 .. 24) {
      my $row_data = substr $data, $rowsize * $row + $colsize * $column, $colsize;
      $column_data .= $row_data;
    }
    $column_data{$component} = $column_data;
  }
  return \%column_data;
}

1;

__END__