package Graphics::Palette;

=head1 NAME

Graphics::Palette - Perl interface for color palette conversions

=head2 SYNOPSIS

use Graphics::Palette qw(get_c64_color);

my $color = get_c64_color('RGB' => '33CC77');
my $color = get_rgb_color('C64' => 0x0C);

my $color = get_c64_color('RGB' => '433900', 'OUTPUT' => 'C64', 'DEBUG' => 1); # $color = '09'

=cut

use bytes;

use strict;
use warnings;

use Carp;
use Data::Dumper;
use Math::VectorReal;

require Exporter;
our @ISA = qw (Exporter);

our @EXPORT_OK   = qw (get_c64_color get_rgb_color);
our %EXPORT_TAGS = ('get_c64_color' => [qw (get_c64_color)], 'get_rgb_color' => [qw (get_rgb_color)]);

local $| = 1;

our @c64_palette = (
    { 'RED' => 0x00, 'GREEN' => 0x00, 'BLUE' => 0x00, 'NAME' => 'BLACK',       },
    { 'RED' => 0xFF, 'GREEN' => 0xFF, 'BLUE' => 0xFF, 'NAME' => 'WHITE',       },
    { 'RED' => 0x68, 'GREEN' => 0x37, 'BLUE' => 0x2B, 'NAME' => 'RED',         },
    { 'RED' => 0x70, 'GREEN' => 0xA4, 'BLUE' => 0xB2, 'NAME' => 'CYAN',        },
    { 'RED' => 0x6F, 'GREEN' => 0x3D, 'BLUE' => 0x86, 'NAME' => 'PURPLE',      },
    { 'RED' => 0x58, 'GREEN' => 0x8D, 'BLUE' => 0x43, 'NAME' => 'GREEN',       },
    { 'RED' => 0x35, 'GREEN' => 0x28, 'BLUE' => 0x79, 'NAME' => 'BLUE',        },
    { 'RED' => 0xB8, 'GREEN' => 0xC7, 'BLUE' => 0x6F, 'NAME' => 'YELLOW',      },
    { 'RED' => 0x6F, 'GREEN' => 0x4F, 'BLUE' => 0x25, 'NAME' => 'ORANGE',      },
    { 'RED' => 0x43, 'GREEN' => 0x39, 'BLUE' => 0x00, 'NAME' => 'BROWN',       },
    { 'RED' => 0x9A, 'GREEN' => 0x67, 'BLUE' => 0x59, 'NAME' => 'LIGHT RED',   },
    { 'RED' => 0x44, 'GREEN' => 0x44, 'BLUE' => 0x44, 'NAME' => 'DARK GREY',   },
    { 'RED' => 0x6C, 'GREEN' => 0x6C, 'BLUE' => 0x6C, 'NAME' => 'GREY',        },
    { 'RED' => 0x9A, 'GREEN' => 0xD2, 'BLUE' => 0x84, 'NAME' => 'LIGHT GREEN', },
    { 'RED' => 0x6C, 'GREEN' => 0x5E, 'BLUE' => 0xB5, 'NAME' => 'LIGHT BLUE',  },
    { 'RED' => 0x95, 'GREEN' => 0x95, 'BLUE' => 0x95, 'NAME' => 'LIGHT GREY',  },
);

sub new {
    my $this  = shift;
    my $class = ref($this) || $this;
    my $self  = {};
    bless $self, $class;
    $self->__initialize( @_ );
    return $self;
}

sub __initialize {
    my $self       = shift;
    my $parameters = {@_};
    $self->debug($parameters->{'DEBUG'});
}

sub DEBUG {
    my $self = shift;
    my $value = shift;
    if (defined $value) {
        $self->{'DEBUG'} = $value;
    }
    return $self->{'DEBUG'};
}

sub get_c64_color {
    my $params = {@_};
    my $rgb    = $params->{'RGB'};
    my $debug  = $params->{'DEBUG'};
    my $output = $params->{'OUTPUT'} || 'PC';
    Carp::croak q{Unable to get C64 color for undefined 'RGB' parameter value within "get_c64_color" subroutine call} unless defined $rgb;
    my ($red, $green, $blue) = map { hex } ($rgb =~ m/([0-9A-F]{2})/ig);
    my $smallest_delta = 0xFFFF;
    my $final_color;
    for (my $i = 0; $i < @c64_palette; $i++) {
        my $c64_color = $c64_palette[$i];
        my $c64_red   = $c64_color->{'RED'};
        my $c64_green = $c64_color->{'GREEN'};
        my $c64_blue  = $c64_color->{'BLUE'};
        # Calculate the distance between both of these color vectors:
        my $a = vector($red, $green, $blue);
        my $b = vector($c64_red, $c64_green, $c64_blue);
        my $c = $a - $b;
        my $delta = $c->length;
        if ($delta < $smallest_delta) {
            if ($output eq 'C64') {
                $final_color = sprintf "%02X", $i;
            }
            else {
                $final_color = sprintf "%02X%02X%02X", $c64_red, $c64_green, $c64_blue;
            }
            $smallest_delta = $delta;
        }
    }
    my $debug_final_color = $output eq 'C64' ? "\$${final_color}" : "#${final_color}";
    print "[debug] rgb2c64 color conversion completed: #${rgb} => ${debug_final_color}\n" if $debug;
    return $final_color;
}

sub get_rgb_color {
    my $params = {@_};
    my $c64    = $params->{'C64'};
    $c64 &= 0x0f;
    my $rgb_color = $c64_palette[$c64];
    return $rgb_color;
}

1;
