package Graphics::Transform;

=head1 NAME

Graphics::Transform - Miscellaneous graphic conversion routines

=head1 SYNOPSIS

  use Graphics::Transform qw(:all);

  # Convert raw bitmap pixels data into charset with an accompanying screen map:
  my ($fonts, $screen_map) = convert_plain_bitmap_to_fonts(
    data => $data,
    rows => $rows,
    cols => $cols,
  );

  # Convert raw bitmap pixels data into charset with an accompanying screen map, starting at a different initial font:
  my ($fonts, $screen_map) = convert_plain_bitmap_to_fonts(
    data => $data,
    rows => $rows,
    cols => $cols,
    offset => $offset,
  );

  # Convert raw bitmap pixels data into charset with an accompanying screen map, reusing an already existing charset:
  my ($fonts, $screen_map) = convert_plain_bitmap_to_fonts(
    data => $data,
    rows => $rows,
    cols => $cols,
    reusable_fonts => $charset,
  );

  # Convert raw bitmap pixels data into sprites:
  my $sprites_data = convert_plain_bitmap_to_sprites(
    data => $data,
    rows => $rows,
    cols => $cols,
  );

=head1 FUNCTIONS

=cut

use utf8;
use bytes;
use strict;
use warnings;

use base qw(Exporter);
our %EXPORT_TAGS = ();
$EXPORT_TAGS{gfx2fonts} = [ qw(&convert_plain_bitmap_to_fonts) ];
$EXPORT_TAGS{gfx2sprites} = [ qw(&convert_plain_bitmap_to_sprites) ];
$EXPORT_TAGS{all} = [ @{ $EXPORT_TAGS{gfx2fonts} }, @{ $EXPORT_TAGS{gfx2sprites} } ];
our @EXPORT_OK = @{ $EXPORT_TAGS{all} };
our @EXPORT = qw();

use Data::Dumper;
use List::Util qw(max);
use Storable qw(dclone);

=head2 convert_plain_bitmap_to_fonts

Convert raw bitmap pixels data into charset with an accompanying screen map, reusing an already existing charset:

  my ($fonts, $screen_map) = convert_plain_bitmap_to_fonts(
    data => $data,
    rows => $rows,
    cols => $cols,
    reusable_fonts => $charset,
  );

Screen map uses provided charset as well as newly generated fonts. Returned fonts consist of newly generated fonts appended to an original charset since they are both required to properly display a target image.

=cut

sub convert_plain_bitmap_to_fonts {
  my (%args) = @_;

  my $cols   = $args{cols}   || 0x19;
  my $data   = $args{data}   || [];
  my $offset = $args{offset} || 0x00;
  my $rows   = $args{rows}   || 0x28;

  my $reusable_fonts = $args{reusable_fonts} || [];

  my @fonts_data = @{ dclone $reusable_fonts };
  my @fonts_index;

  $offset += scalar @fonts_data;

  # Split bitmap data into fonts and screen map:
  while ($data =~ m/(.{8})/gs) {
    my $font = $+;
    push @fonts_data, $font;
    push @fonts_index, $offset++;
  }

  # Find duplicate fonts in a generated charset:
  my @duplicates;
  for my $index ( 0x00 .. $#fonts_data ) {
    my $test_font = $fonts_data[$index];
    # Check all the remaining fonts, whether they match currently tested font:
    for (my $i = $index + 1; $i < @fonts_data; $i++ ) {
      # Do not check if current font is already a duplicate of another font:
      next if defined $duplicates[$i];
      if ($test_font eq $fonts_data[$i]) {
        # Font at offset $i is identical to font at offset $index:
        $duplicates[$i] = $index;
      }
    }
  }

  # Eliminate duplicate fonts from generated charset:
  for my $index ( 0x00 .. $#duplicates) {
    my $original = $duplicates[$index];
    next unless defined $original;
    # Insert original font in a place of a duplicate:
    for (my $i = 0; $i < @fonts_index; $i++) {
      if ($fonts_index[$i] == $index) {
        $fonts_index[$i] = $original;
      }
    }
    # Remove duplicate font data:
    $fonts_data[$index] = undef;
  }

  # Remove all empty chars from end of fonts data automatically:
  pop @fonts_data while not defined $fonts_data[-1];

  # Cleanup empty spots in fonts data and screen map:
  while (grep { not defined } @fonts_data) {
    # Find the first empty slot:
    my ($new_index) = map { defined $fonts_data[$_] ? () : $_ } ( 0x00 .. $#fonts_data );
    # Get the last non-empty font:
    my $old_index = $#fonts_data;
    # Move non-empty font into empty slot:
    $fonts_data[$new_index] = $fonts_data[$old_index];
    pop @fonts_data;
    # Change all fonts in screen map to point to the new font data offset:
    for (my $i = 0; $i < @fonts_index; $i++) {
      if ($fonts_index[$i] == $old_index) {
        $fonts_index[$i] = $new_index;
      }
    }
    # Remove all empty chars from end of fonts data automatically:
    pop @fonts_data while not defined $fonts_data[-1];
  }

  # Throw an exception when number of used fonts exceeds available memory:
  my $max_font_index = max @fonts_index;
  if ($max_font_index > 0xff) {
    die sprintf 'Insufficient memory to generate %d fonts', $max_font_index + 1;
  }

  # Place fonts on screen map according to provided image dimensions:
  my @screen_map;
  while (my @row = splice @fonts_index, 0x00, $cols) {
    push @screen_map, join '', map { chr } @row;
  }

  return (\@fonts_data, \@screen_map);
}

=head2 convert_plain_bitmap_to_sprites

Convert raw bitmap pixels data into sprites:

  my $sprites_data = convert_plain_bitmap_to_sprites(
    data => $data,
    rows => $rows,
    cols => $cols,
  );

=cut

sub convert_plain_bitmap_to_sprites {
  my (%args) = @_;

  my $cols = $args{cols} || 0x01;
  my $data = $args{data} || [];
  my $rows = $args{rows} || 0x01;

  my @sprites_data;

  for (my $spr_y = 0x00; $spr_y < $rows; $spr_y++) {
    for (my $spr_x = 0x00; $spr_x < $cols; $spr_x++) {
      my @sprite_data;
      for (my $j = 0x00; $j < 0x15; $j++) {
        for (my $i = 0x00; $i < 0x03; $i++) {
          my $x = $spr_x * 0x03 + $i;
          my $y = $spr_y * 0x15 + $j;
          my $offset = ($y >> 0x03) * 0x0140 + ($y & 0x07) + $x * 0x08;
          my $byte = ord substr $data, $offset, 0x01;
          push @sprite_data, $byte;
        }
      }
      push @sprite_data, 0x00;
      push @sprites_data, \@sprite_data;
    }
  }

  return \@sprites_data;
}

1;

__END__