package Memory::Address;

=head1 NAME

Memory::Address - Handling 16-bit memory address calculations and conversions

=head1 SYNOPSIS

  use Memory::Address qw(:all);

  # Convert numeric memory address to lo/hi bytes in a scalar context:
  my $data = address_to_lo_hi_bytes(0xe000);

  # Convert numeric memory address to lo/hi bytes in an array context:
  my @data = address_to_lo_hi_bytes(0xe000);

  # Get file loading address from the initial two bytes of its contents:
  my $loading_address = get_file_loading_address('filename.prg');

  # Convert numeric memory address to a hexadecimal string representation:
  my $text = address_to_hex_string(0xe000);

=cut

use bytes;
use strict;
use utf8;
use warnings;

use base qw(Exporter);
our %EXPORT_TAGS = ();
$EXPORT_TAGS{'convert'} = [ qw(&address_to_lo_hi_bytes &lo_hi_bytes_to_address) ];
$EXPORT_TAGS{'file'} = [ qw(&get_file_loading_address) ];
$EXPORT_TAGS{'format'} = [ qw(&address_to_hex_string) ];
$EXPORT_TAGS{'all'} = [ @{ $EXPORT_TAGS{'convert'} }, @{ $EXPORT_TAGS{'file'} }, @{ $EXPORT_TAGS{'format'} } ];
our @EXPORT_OK = @{ $EXPORT_TAGS{'all'} };
our @EXPORT = qw();

use Data::Dumper;
use IO::File;

sub address_to_hex_string {
  my ($addr) = @_;
  my @data = address_to_lo_hi_bytes($addr);
  my $text = sprintf '$%02x%02x', ord ($data[1]), ord ($data[0]);
  return $text;
}

sub address_to_lo_hi_bytes {
  my ($addr) = @_;
  return unless defined wantarray;
  my $lo = $addr & 0x00ff;
  my $hi = ($addr >> 0x08) & 0x00ff;
  my @bytes = map { chr } ($lo, $hi);
  my $bytes = join '', @bytes;
  return wantarray ? @bytes : $bytes;
}

sub lo_hi_bytes_to_address {
  my (@bytes) = @_;
  @bytes = split //, $bytes[0] if scalar @bytes == 1;
  my $addr = ord ($bytes[0]) + ord ($bytes[1]) * 0x0100;
  return $addr;
}

sub get_file_loading_address {
  my ($file) = @_;
  my $fh = new IO::File $file, 'r';
  if (defined $fh) {
    my $bytes;
    sysread $fh, $bytes, 0x0002;
    $fh->close;
    my $addr = lo_hi_bytes_to_address($bytes);
    return $addr;
  }
  return;
}

1;
