package Misc::Util;

=head1 NAME

Misc::Util - miscellaneous file utilities

=head1 SYNOPSIS

use Misc::Util qw(:file);

crash_on_file_exists($file);
crash_on_file_not_exists($file);
crash_on_file_not_readable($file);

my ($data_aref) = read_file_raw_data($file);
write_file_raw_data($file, $data_aref);

my ($loadaddr, $data_aref) = read_c64_file_raw_data($file);
write_c64_file_raw_data($file, $data_aref, $loadaddr);

modify_c64_file_loading_address($filename, $address);

use Misc::Util qw(:addr);

my ($loadaddr_lo, $loadaddr_hi) = memo_addr_to_two_bytes($loadaddr);
my ($loadaddr) = two_bytes_to_memo_addr($loadaddr_lo, $loadaddr_hi);

use Misc::Util qw(:rand);

my $randstr = generate_random_string($length);

=cut

use strict;
use warnings;

use Carp qw(verbose);
use Data::Dumper;

require Exporter;
our @ISA = qw(Exporter);

our @EXPORT_OK   = qw(crash_on_file_exists crash_on_file_not_exists crash_on_file_not_readable generate_random_string get_file_type memo_addr_to_two_bytes modify_c64_file_loading_address read_c64_file_raw_data read_file_raw_data two_bytes_to_memo_addr write_c64_file_raw_data write_file_raw_data);
our %EXPORT_TAGS = ('addr' => [qw(memo_addr_to_two_bytes two_bytes_to_memo_addr)], 'file' => [qw(crash_on_file_exists crash_on_file_not_exists crash_on_file_not_readable get_file_type modify_c64_file_loading_address read_c64_file_raw_data read_file_raw_data write_c64_file_raw_data write_file_raw_data)], 'rand' => [qw(generate_random_string)]);

local $| = 1;

##########
## ADDR ##
##########

sub memo_addr_to_two_bytes {
    my $memo_addr = shift;
    my $address_as_str = sprintf "%04x", $memo_addr;
    my ($memo_addr_hi, $memo_addr_lo) = ($address_as_str =~ m/([0-9a-f]{2})/gi);
    $memo_addr_hi = chr hex $memo_addr_hi;
    $memo_addr_lo = chr hex $memo_addr_lo;
    return ($memo_addr_lo, $memo_addr_hi);
}

sub two_bytes_to_memo_addr {
    my $memo_addr_lo = shift;
    my $memo_addr_hi = shift;
    my $address_as_num = ord ($memo_addr_lo) + ord ($memo_addr_hi) * 0x0100;
    return $address_as_num;
}

##########
## FILE ##
##########

sub modify_c64_file_loading_address {
    my $file = shift;
    my $new_loadaddr = shift;
    my ($old_loadaddr, $data_aref) = read_c64_file_raw_data($file);
    my $temp_file = generate_random_string(16);
    write_c64_file_raw_data($temp_file, $data_aref, $new_loadaddr);
    unlink ($file);
    rename ($temp_file, $file);
}

sub read_c64_file_raw_data {
    my $file = shift;
    my ($data_aref) = read_file_raw_data($file);
    my ($loadaddr_lo, $loadaddr_hi) = splice (@{$data_aref}, 0, 2);
    my $loadaddr = two_bytes_to_memo_addr($loadaddr_lo, $loadaddr_hi);
    return ($loadaddr, $data_aref);
}

sub read_file_raw_data {
    my $file = shift;
    crash_on_file_not_readable($file);
    open my $fh, '<', $file or Carp::croak "File open \"${file}\" operation failed";
    binmode $fh, ":bytes";
    my (@buffer, $data, $n) = ();
    push @buffer, $data while  ($n = read $fh, $data, 1) != 0;
    close $fh or Carp::croak "File close \"${file}\" operation failed";
    return \@buffer;
}

sub write_c64_file_raw_data {
    my $file = shift;
    my $data_aref = shift;
    my $loadaddr = shift;
    my @data = @{$data_aref};
    my @loadaddr_bytes = memo_addr_to_two_bytes($loadaddr);
    unshift @data, @loadaddr_bytes;
    write_file_raw_data($file, \@data);
}

sub write_file_raw_data {
    my $file = shift;
    my $data_ref = shift;
    crash_on_file_exists($file);
    open my $fh, '>', $file or Carp::croak "File open \"${file}\" operation failed";
    binmode $fh, ":bytes";
    foreach my $byte (@{$data_ref}) {
        Carp::croak 'An undefined value cannot be written to a file handle' unless defined $byte;
        syswrite $fh, $byte, 1;
    }
    close $fh or Carp::croak "File close \"${file}\" operation failed";
}

sub crash_on_file_exists {
    my $file = shift;
    if (not defined $file) {
        Carp::croak 'No file name provided';
    }
    if (-e $file) {
        Carp::croak "File \"${file}\" already exists";
    }
}

sub crash_on_file_not_exists {
    my $file = shift;
    if (not defined $file) {
        Carp::croak 'No file name provided';
    }
    if (not -e $file) {
        Carp::croak "File \"${file}\" does not exist";
    }
}

sub crash_on_file_not_readable {
    my $file = shift;
    if (not defined $file) {
        Carp::croak 'No file name provided';
    }
    if (not -r $file) {
        Carp::croak "File \"${file}\" cannot be read from";
    }
}

sub get_file_type {
    my $file = shift;
    my $type = '';
    if (-f $file) {
        $type = 'plain file';
    }
    elsif (-d $file) {
        $type = 'directory';
    }
    elsif (-l $file) {
        $type = 'symbolic link';
    }
    elsif (-p $file) {
        $type = 'named pipe (FIFO)';
    }
    elsif (-S $file) {
        $type = 'socket';
    }
    elsif (-b $file) {
        $type = 'block special file';
    }
    elsif (-c $file) {
        $type = 'character special file';
    }
    elsif (-t $file) {
        $type = 'filehandle opened to a tty';
    }
    else {
        $type = 'unknown';
    }
    return $type;
}

##########
## RAND ##
##########

sub generate_random_string {
    my $length_of_random_string = shift;
    my @chars = ('a' .. 'z', 'A' .. 'Z', '0' .. '9', '_');
    my $random_string = '';
    foreach (1 .. $length_of_random_string) {
        $random_string .= $chars[rand @chars];
    }
    return $random_string;
}

1;
