package Source::Data;

use bytes;
use strict;
use warnings;

use base qw(Exporter);
our %EXPORT_TAGS = ();
$EXPORT_TAGS{'all'} = [ qw(&generate_source_data) ];
our @EXPORT_OK = @{ $EXPORT_TAGS{'all'} };
our @EXPORT = qw();

use Carp qw(croak);
use Data::Dumper;
use List::Util qw(max);
use Text::Convert::PETSCII qw/:all/;

sub generate_source_data {
    my %params = @_;

    my $lookup_table           = $params{data};
    my $table_name             = $params{name};
    my $base_address           = $params{base_address} || '';
    my $values_per_line        = $params{values_per_line} || 8;
    my $values_dump            = $values_per_line == 1 ? q{SINGLE} : q{MULTIPLE};
    my $additional_indentation = $params{additional_indentation} || 0;
    my $expected_indentation   = $params{expected_indentation} || 0;
    my $calculate_hex_values   = $params{calculate_hex_values} || 0;
    my $include_comments       = $params{include_comments} || 0;
    my $lo_and_hi_tables_gen   = $params{lo_and_hi_tables_gen} || 0;
    my $data_type              = $params{data_type} || 'bytes'; # bytes/petscii

    if ($expected_indentation) {
        if (length ($table_name) < $expected_indentation) {
            $additional_indentation = $expected_indentation - length ($table_name) - 1;
        }
        if ($lo_and_hi_tables_gen && $values_dump eq q{SINGLE}) {
            $additional_indentation -= 3; # because of "_lo" and "_hi" suffixes!
        }
    }

    my @values_presufixes = $lo_and_hi_tables_gen
        ? ({value_prefix => q{<}, table_name_suffix => q{_lo}}, {value_prefix => q{>}, table_name_suffix => q{_hi}})
        : ({value_prefix => q{}, table_name_suffix => q{}});

    my $all_table_dumps;

    for my $presufix (@values_presufixes) {
        my $table_name_suffix = $presufix->{table_name_suffix};
        my $value_prefix = $presufix->{value_prefix};

        my $lookup_table_dump = $table_name . $table_name_suffix . q{ };
        my $initial_label_length = length $lookup_table_dump;

        if ($values_dump eq q{SINGLE}) {
            $lookup_table_dump .= q { } x $additional_indentation;

            for (my $x = 0; $x < @{$lookup_table}; $x++) {
                my $raw_value = $calculate_hex_values ? sprintf (q{$%02x}, $lookup_table->[$x]) : $lookup_table->[$x];

                my $value = sprintf q{%s(%s%s)}, $value_prefix, $base_address, $raw_value;
                unless ($x == 0) {
                    if ($expected_indentation) {
                        $lookup_table_dump .= q{ } x ($expected_indentation - $additional_indentation);
                    }
                    $lookup_table_dump .= q{ } x $additional_indentation;
                }
                $lookup_table_dump .= sprintf qq{%s %s\n}, _format_dump_opcode($data_type), $value;
            }
        } elsif ($values_dump eq q{MULTIPLE}) {
            for (my $x = 0; $x < @{$lookup_table}; $x = $x + $values_per_line) {
                my @line;
                for (my $i = 0; $i < $values_per_line; $i++) {
                    next unless defined $lookup_table->[$x + $i];

                    my $raw_value = $calculate_hex_values ? sprintf (q{$%02x}, $lookup_table->[$x + $i]) : _format_value($lookup_table->[$x + $i], $data_type);
                    $raw_value .= q{)} if $base_address;

                    my $value = sprintf q{%s%s%s}, $value_prefix, $base_address, $raw_value;
                    push @line, $value;
                }
                if ($expected_indentation) {
                    if ($x == 0) {
                        $lookup_table_dump .= q{ } x ($expected_indentation - $initial_label_length);
                    }
                    else {
                        $lookup_table_dump .= q{ } x ($expected_indentation);
                    }
                }
                $lookup_table_dump .= sprintf qq{%s %s\n}, _format_dump_opcode($data_type), join (q{,}, @line);
            }
        }
        else {
            die sprintf q{Invalid values dump requested: "%s"}, $values_dump;
        }

        if ($include_comments) {
            my $comment_length = max map { length } split "\n", $lookup_table_dump;
            my $comment_text = ';' . '-' x ($comment_length - 1) . "\n";

            $lookup_table_dump = $comment_text . $lookup_table_dump;
        }

        $all_table_dumps .= $lookup_table_dump;
    }

    my $comment_length = max map { length } split "\n", $all_table_dumps;
    my $comment_text = ';' . '-' x ($comment_length - 1) . "\n";

    $all_table_dumps = $all_table_dumps . $comment_text;

    return $all_table_dumps;
}

sub _format_value {
  my ($value, $data_type) = @_;
  if ($data_type eq 'bytes') {
    return $value;
  }
  if ($data_type eq 'petscii') {
    # my $petscii = ascii_to_petscii($value);
    # return sprintf '$%02x', ord $petscii;
    return sprintf '"%s"', $value;
  }
  die "Unrecognized data format: $data_type";
}

sub _format_dump_opcode {
  my ($data_type) = @_;
  if ($data_type eq 'bytes') {
    return '.db';
  }
  if ($data_type eq 'petscii') {
    return '.dp';
  }
  die "Unrecognized data format: $data_type";
}

1;

__END__